"""
This **module** gives the basic functions for fep calculations
"""
from ...helper import source, set_global_alternative_names, Xdict, Xopen, Xprint
from ..base import lj_base, exclude_base, bond_base, angle_base, dihedral_base, nb14_extra_base

source("....")

lj_type = lj_base.LJType

lj_type.New_From_String("""name A  B
ZERO_LJ_ATOM-ZERO_LJ_ATOM  0  0
""")

AtomType.Add_Property({"LJtypeB": str})
AtomType.Add_Property({"subsys": int})


def _find_common_forces(forcetype, aforces, bforces, mol_b2mol_a):
    """

    :param forcetype:
    :param aforces:
    :param bforces:
    :param mol_b2mol_a:
    :return:
    """
    toret = []
    temp_map = Xdict()
    temp_map2 = Xdict()
    for force in bforces:
        temp_map2[force] = True
        for fatoms in forcetype.Same_Force(force.atoms):
            temp_map["-".join(list(map(lambda atom: str(mol_b2mol_a[atom]), fatoms)))] = force
    for force in aforces:
        tofind = "-".join(list(map(str, force.atoms)))
        bforce = temp_map.get(tofind, None)
        toret.append([force, bforce])
        temp_map2[bforce] = False
    for force in bforces:
        if temp_map2[force]:
            toret.append([None, force])
    return toret


FEP_BONDED_FORCE_MERGE_RULE = Xdict()
TINY = 1e-10 / 18.2223


# pylint: disable=unused-argument
def _nb14_extra_merge_rule(mol_r, mol_a, mol_b, forcetype, rforces, bforces, lambda_, mol_r2mol_a, mol_a2mol_r,
                           mol_r2mol_b, mol_b2mol_r):
    """

    :param mol_r:
    :param mol_a:
    :param mol_b:
    :param forcetype:
    :param rforces:
    :param bforces:
    :param lambda_:
    :param mol_r2mol_a:
    :param mol_a2mol_r:
    :param mol_r2mol_b:
    :param mol_b2mol_r:
    :return:
    """
    forcepair = _find_common_forces(forcetype, rforces, bforces, mol_b2mol_r)
    for f_r, f_b in forcepair:
        if f_b is None:
            temp_charge0 = f_r.atoms[0].charge if abs(f_r.atoms[0].charge) > TINY else TINY
            temp_charge1 = f_r.atoms[1].charge if abs(f_r.atoms[1].charge) > TINY else TINY

            if f_r.nb14_ee_factor is not None:
                nb14_ee_factor = f_r.nb14_ee_factor
            else:
                nb14_ee_factor = f_r.kee * mol_r2mol_a[f_r.atoms[0]].charge * \
                                 mol_r2mol_a[f_r.atoms[1]].charge
            f_r.kee = nb14_ee_factor / temp_charge0 / temp_charge1

            f_r.kee *= 1 - lambda_
            f_r.A *= 1 - lambda_
            f_r.B *= 1 - lambda_
        elif f_r is None:
            f_r = nb14_extra_base.NB14Type.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)

            temp_charge0 = f_r.atoms[0].charge if abs(f_r.atoms[0].charge) > TINY else TINY
            temp_charge1 = f_r.atoms[1].charge if abs(f_r.atoms[1].charge) > TINY else TINY

            if f_b.nb14_ee_factor is not None:
                nb14_ee_factor = f_b.nb14_ee_factor
            else:
                nb14_ee_factor = f_b.kee * f_b.atoms[0].charge * \
                                 f_b.atoms[1].charge
            f_r.kee = nb14_ee_factor / temp_charge0 / temp_charge1

            f_r.A = f_b.A * lambda_
            f_r.B = f_b.B * lambda_
            f_r.kee = f_b.kee * lambda_
            mol_r.Add_Bonded_Force(f_r)
        else:
            temp_charge0 = f_r.atoms[0].charge if abs(f_r.atoms[0].charge) > TINY else TINY
            temp_charge1 = f_r.atoms[1].charge if abs(f_r.atoms[1].charge) > TINY else TINY

            if f_r.nb14_ee_factor is not None:
                nb14_ee_factor = f_r.nb14_ee_factor
            else:
                nb14_ee_factor = f_r.kee * mol_r2mol_a[f_r.atoms[0]].charge * \
                                 mol_r2mol_a[f_r.atoms[1]].charge
            f_r.kee = nb14_ee_factor / temp_charge0 / temp_charge1

            if f_b.nb14_ee_factor is not None:
                nb14_ee_factor = f_b.nb14_ee_factor
            else:
                nb14_ee_factor = f_b.kee * f_b.atoms[0].charge * \
                                 f_b.atoms[1].charge
            kee = nb14_ee_factor / temp_charge0 / temp_charge1

            f_r.kee = f_r.kee * (1 - lambda_) + kee * lambda_
            f_r.A = f_r.A * (1 - lambda_) + f_b.A * lambda_
            f_r.B = f_r.B * (1 - lambda_) + f_b.B * lambda_


FEP_BONDED_FORCE_MERGE_RULE["nb14_extra"] = {"lambda_name": "dihedral", "merge_function": _nb14_extra_merge_rule}


def _bond_base_merge_rule(mol_r, mol_a, mol_b, forcetype, rforces, bforces, lambda_,
                          mol_r2mol_a, mol_a2mol_r, mol_r2mol_b, mol_b2mol_r):
    """

    :param mol_r:
    :param mol_a:
    :param mol_b:
    :param forcetype:
    :param rforces:
    :param bforces:
    :param lambda_:
    :param mol_r2mol_a:
    :param mol_a2mol_r:
    :param mol_r2mol_b:
    :param mol_b2mol_r:
    :return:
    """
    forcepair = _find_common_forces(forcetype, rforces, bforces, mol_b2mol_r)
    for f_r, f_b in forcepair:
        if f_b is None:
            mol_r.bonded_forces["bond_base"].remove(f_r)
            f_r.from_AorB = 0
            mol_r.Add_Bonded_Force(f_r, "soft_bond_base")
        elif f_r is None:
            f_r = bond_base.BondType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r.k = f_b.k
            f_r.b = f_b.b
            f_r.from_AorB = 1
            mol_r.Add_Bonded_Force(f_r, "soft_bond_base")
        elif abs(f_r.b - f_b.b) < 1e-5:
            f_r.k = f_r.k * (1 - lambda_) + f_b.k * lambda_
        else:
            f_r2 = bond_base.BondType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r2.k = f_b.k * lambda_
            f_r.b = f_b.b
            f_r.k *= (1 - lambda_)
            mol_r.Add_Bonded_Force(f_r2)


FEP_BONDED_FORCE_MERGE_RULE["bond"] = {"lambda_name": "bond_base", "merge_function": _bond_base_merge_rule}


def _angle_merge_rule(mol_r, mol_a, mol_b, forcetype, rforces, bforces, lambda_, mol_r2mol_a, mol_a2mol_r, mol_r2mol_b,
                      mol_b2mol_r):
    """

    :param mol_r:
    :param mol_a:
    :param mol_b:
    :param forcetype:
    :param rforces:
    :param bforces:
    :param lambda_:
    :param mol_r2mol_a:
    :param mol_a2mol_r:
    :param mol_r2mol_b:
    :param mol_b2mol_r:
    :return:
    """
    forcepair = _find_common_forces(forcetype, rforces, bforces, mol_b2mol_r)
    for f_r, f_b in forcepair:
        if f_b is None:
            f_r.k *= 1 - lambda_
        elif f_r is None:
            f_r = angle_base.AngleType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r.k *= lambda_
            f_r.b = f_b.b
            mol_r.Add_Bonded_Force(f_r)
        elif abs(f_r.b - f_b.b) < 1e-5:
            f_r.k = f_r.k * (1 - lambda_) + f_b.k * lambda_
        else:
            f_r2 = angle_base.AngleType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r2.k = f_b.k * lambda_
            f_r2.b = f_b.b
            f_r.k *= (1 - lambda_)
            mol_r.Add_Bonded_Force(f_r2)


FEP_BONDED_FORCE_MERGE_RULE["angle"] = {"lambda_name": "angle", "merge_function": _angle_merge_rule}


def _dihedral_merge_rule(mol_r, mol_a, mol_b, forcetype, rforces, bforces, lambda_, mol_r2mol_a, mol_a2mol_r,
                         mol_r2mol_b, mol_b2mol_r):
    """

    :param mol_r:
    :param mol_a:
    :param mol_b:
    :param forcetype:
    :param rforces:
    :param bforces:
    :param lambda_:
    :param mol_r2mol_a:
    :param mol_a2mol_r:
    :param mol_r2mol_b:
    :param mol_b2mol_r:
    :return:
    """
    forcepair = _find_common_forces(forcetype, rforces, bforces, mol_b2mol_r)
    for f_r, f_b in forcepair:
        if f_b is None:
            for i in range(f_r.multiple_numbers):
                f_r.ks[i] *= 1 - lambda_
        elif f_r is None:
            f_r2 = dihedral_base.ProperType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r2.multiple_numbers = f_b.multiple_numbers
            for i in range(f_b.multiple_numbers):
                f_r2.ks.append(f_b.ks[i] * lambda_)
                f_r2.phi0s.append(f_b.phi0s[i])
                f_r2.periodicitys.append(f_b.periodicitys[i])
            mol_r.Add_Bonded_Force(f_r2)
        else:
            sameforce = f_r.multiple_numbers == f_b.multiple_numbers
            if sameforce:
                check_map = Xdict()
                for i in range(f_b.multiple_numbers):
                    check_map[f_b.periodicitys[i]] = f_b.phi0s[i]
                for i in range(f_r.multiple_numbers):
                    if abs(check_map.get(f_r.periodicitys[i], float("Inf")) - f_r.phi0s[i]) > 1e-5:
                        sameforce = False
                        break
            if sameforce:
                for i in range(f_r.multiple_numbers):
                    f_r.ks[i] = f_r.ks[i] * (1 - lambda_) + f_b.ks[i] * lambda_
            else:
                for i in range(f_r.multiple_numbers):
                    f_r.ks[i] *= (1 - lambda_)
                f_r2 = dihedral_base.ProperType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type,
                                                       f_b.name)
                f_r2.multiple_numbers = f_b.multiple_numbers
                for i in range(f_b.multiple_numbers):
                    f_r2.ks.append(f_b.ks[i] * lambda_)
                    f_r2.phi0s.append(f_b.phi0s[i])
                    f_r2.periodicitys.append(f_b.periodicitys[i])
                mol_r.Add_Bonded_Force(f_r2)


FEP_BONDED_FORCE_MERGE_RULE["dihedral"] = {"lambda_name": "dihedral", "merge_function": _dihedral_merge_rule}


def _improper_merge_rule(mol_r, mol_a, mol_b, forcetype, rforces, bforces, lambda_, mol_r2mol_a, mol_a2mol_r,
                         mol_r2mol_b, mol_b2mol_r):
    """

    :param mol_r:
    :param mol_a:
    :param mol_b:
    :param forcetype:
    :param rforces:
    :param bforces:
    :param lambda_:
    :param mol_r2mol_a:
    :param mol_a2mol_r:
    :param mol_r2mol_b:
    :param mol_b2mol_r:
    :return:
    """
    forcepair = _find_common_forces(forcetype, rforces, bforces, mol_b2mol_r)
    for f_r, f_b in forcepair:
        if f_b is None:
            f_r.k *= 1 - lambda_
        elif f_r is None:
            f_r = dihedral_base.ImproperType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r.k = f_b.k * lambda_
            f_r.phi0 = f_b.phi0
            f_r.periodicity = f_b.periodicity
            mol_r.Add_Bonded_Force(f_r)
        elif abs(f_r.phi0 - f_b.phi0) < 1e-5 and f_r.periodicity == f_b.periodicity:
            f_r.k = f_r.k * (1 - lambda_) + f_b.k * lambda_
        else:
            f_r2 = dihedral_base.ImproperType.entity(list(map(lambda x: mol_b2mol_r[x], f_b.atoms)), f_b.type, f_b.name)
            f_r2.k = f_b.k * lambda_
            f_r2.phi0 = f_b.phi0
            f_r2.periodicity = f_b.periodicity
            f_r.k *= (1 - lambda_)
            mol_r.Add_Bonded_Force(f_r2)


FEP_BONDED_FORCE_MERGE_RULE["improper"] = {"lambda_name": "dihedral", "merge_function": _improper_merge_rule}


def save_hard_core_lj():
    """
    This **function** is used to save hard core lj

    :return: None
    """
    Molecule.Set_Save_SPONGE_Input("LJ")(lj_base.write_zero_lj)
    Molecule.Del_Save_SPONGE_Input("LJ_soft_core")
    Molecule.Del_Save_SPONGE_Input("subsys_division")


def save_soft_core_lj():
    """
    This **function** is used to save soft core lj

    :return: None
    """
    Molecule.Del_Save_SPONGE_Input("LJ")

    def write_subsys_division(self):
        towrite = "%d\n" % len(self.atoms)
        for atom in self.atoms:
            if getattr(atom, "subsys", None) is None:
                towrite += "%d\n" % 0
            else:
                towrite += "%d\n" % atom.subsys
        return towrite

    Molecule.Set_Save_SPONGE_Input("subsys_division")(write_subsys_division)

    def write_soft_lj(self):
        """

        :param self:
        :return:
        """
        lj_types = []
        lj_typemap = Xdict()
        lj_typeb = []
        lj_typemapb = Xdict()
        for atom in self.atoms:
            if atom.LJtype not in lj_typemap.keys():
                lj_typemap[atom.LJtype] = len(lj_types)
                lj_types.append(atom.LJtype)
            if atom.LJtypeB is None:
                atom.LJtypeB = atom.LJtype
            if atom.LJtypeB not in lj_typemapb.keys():
                lj_typemapb[atom.LJtypeB] = len(lj_typeb)
                lj_typeb.append(atom.LJtypeB)

        # pylint: disable=protected-access
        lj_as, lj_bs = lj_base._find_ab_lj(lj_types)
        lj_asb, lj_bsb = lj_base._find_ab_lj(lj_typeb)
        checks = lj_base._get_checks(lj_types, lj_as, lj_bs)
        same_type = lj_base._judge_same_type(lj_types, checks)
        real_lj_types = lj_base._get_real_lj(lj_types, same_type)
        real_as, real_bs = lj_base._find_ab_lj(real_lj_types, False)

        checks = lj_base._get_checks(lj_typeb, lj_asb, lj_bsb)
        same_typeb = lj_base._judge_same_type(lj_typeb, checks)
        real_lj_typesb = lj_base._get_real_lj(lj_typeb, same_typeb)
        real_asb, real_bsb = lj_base._find_ab_lj(real_lj_typesb, False)

        towrite = "%d %d %d\n\n" % (len(self.atoms), len(real_lj_types), len(real_lj_typesb))
        count = 0
        for i in range(len(real_lj_types)):
            for _ in range(i + 1):
                towrite += "%16.7e" % real_as[count] + " "
                count += 1
            towrite += "\n"
        towrite += "\n"

        count = 0
        for i in range(len(real_lj_types)):
            for _ in range(i + 1):
                towrite += "%16.7e" % real_bs[count] + " "
                count += 1
            towrite += "\n"
        towrite += "\n"

        count = 0
        for i in range(len(real_lj_typesb)):
            for _ in range(i + 1):
                towrite += "%16.7e" % real_asb[count] + " "
                count += 1
            towrite += "\n"
        towrite += "\n"

        count = 0
        for i in range(len(real_lj_typesb)):
            for _ in range(i + 1):
                towrite += "%16.7e" % real_bsb[count] + " "
                count += 1
            towrite += "\n"

        towrite += "\n"
        towrite += "\n".join(
            ["%d %d" % (same_type[lj_typemap[atom.LJtype]], same_typeb[lj_typemapb[atom.LJtypeB]]) for atom in
             self.atoms])
        return towrite

    Molecule.Set_Save_SPONGE_Input("LJ_soft_core")(write_soft_lj)


def intramolecule_nb_to_nb14(mol_a, perturbating_residues):
    """
    This **function** convert the non bonded interactions to nb14 interactions within the molecule

    :param mol_a: the Molecule instance
    :param perturbating_residues: the residue(s) to be perturbed
    :return:
    """
    if isinstance(perturbating_residues, Residue):
        perturbating_residues = [perturbating_residues]
    build.Build_Bonded_Force(mol_a)
    a_exclude = exclude_base.Exclude.current.Get_Excluded_Atoms(mol_a)
    for residue1 in perturbating_residues:
        for atom_a1 in residue1.atoms:
            for residue2 in perturbating_residues:
                for atom_a2 in residue2.atoms:
                    if atom_a1 == atom_a2:
                        continue
                    if atom_a2 not in a_exclude[atom_a1]:
                        temp_a, temp_b = nb14_extra_base.Get_NB14EXTRA_AB(atom_a1, atom_a2)
                        new_force = nb14_extra_base.NB14Type.entity([atom_a1, atom_a2],
                                                                    nb14_extra_base.NB14Type.get_type("UNKNOWNS"))
                        new_force.A = temp_a
                        new_force.B = temp_b
                        new_force.kee = 1
                        mol_a.Add_Bonded_Force(new_force)

                        atom_a1.Extra_Exclude_Atom(atom_a2)
                        a_exclude[atom_a1].add(atom_a2)
                        a_exclude[atom_a2].add(atom_a1)


def get_free_molecule(mol_a, perturbating_residues, intra_fep=False):
    """
    This **function** makes the molecule to be "free", having no interaction with other molecules

    :param mol_a: the Molecule instance
    :param perturbating_residues: the residues to be perturbed
    :param intra_fep: whether clear intramolecular non bonded interactions
    :return: a new Molecule instance
    """
    if isinstance(perturbating_residues, Residue):
        perturbating_residues = [perturbating_residues]
    build.Build_Bonded_Force(mol_a)
    nb14_extra_base.NB14_To_NB14EXTRA(mol_a)
    mol_b = mol_a.deepcopy()
    mol_a2mol_b = Xdict()
    for i, atom_a in enumerate(mol_a.atoms):
        mol_a2mol_b[atom_a] = mol_b.atoms[i]

    for residue in perturbating_residues:
        for atom_a in residue.atoms:
            atom = mol_a2mol_b[atom_a]
            atom.charge = 0
            atom.LJtype = "ZERO_LJ_ATOM"
            atom.subsys = 1
            atom_a.subsys = 1

    return mol_b


def _correct_residueb_coordinates(residue_a, residue_b, matchmap):
    """

    :param residue_a:
    :param residue_b:
    :param matchmap:
    :return:
    """
    uncertified = set([])
    certified = Xdict()
    for i, atom in enumerate(residue_b.atoms):
        if i in matchmap.keys():
            temp_atom = residue_a.atoms[matchmap[i]]
            certified[atom] = [temp_atom.x, temp_atom.y, temp_atom.z]
        else:
            uncertified.add(i)
    while uncertified:
        movedlist = []
        for i in uncertified:
            atom = residue_b.atoms[i]
            for connected_atom in residue_b.connectivity[atom]:
                if connected_atom in certified.keys():
                    temp_list = [0, 0, 0]
                    temp_list[0] = atom.x - connected_atom.x + certified[connected_atom][0]
                    temp_list[1] = atom.y - connected_atom.y + certified[connected_atom][1]
                    temp_list[2] = atom.z - connected_atom.z + certified[connected_atom][2]
                    certified[atom] = temp_list
                    movedlist.append(i)
                    break
        for i in movedlist:
            uncertified.remove(i)
    for atom, crd in certified.items():
        atom.x, atom.y, atom.z = crd[0], crd[1], crd[2]


def _get_extra_atoms_and_rbmap(restype_ab, residue_type_a, residue_type_b, residue_a,
                               forcopy, matchmap, match_a, match_b):
    """

    :param restype_ab:
    :param residue_type_a:
    :param residue_type_b:
    :param residue_a:
    :param forcopy:
    :param matchmap:
    :param match_a:
    :param match_b:
    :return:
    """
    extra_a = []
    extra_b = []
    rbmap = {value: key for key, value in matchmap.items()}
    for i, atom in enumerate(residue_type_a.atoms):
        atom.x = residue_a.name2atom(atom.name).x
        atom.y = residue_a.name2atom(atom.name).y
        atom.z = residue_a.name2atom(atom.name).z
        if i not in match_a:
            extra_a.append(atom.copied[forcopy])
            extra_a[-1].subsys = 1

    for i, atom in enumerate(residue_type_b.atoms):
        if i not in match_b:
            rbmap[len(restype_ab.atoms)] = i
            restype_ab.Add_Atom(atom.name + "R2", atom.type, atom.x, atom.y, atom.z)
            atom.copied[forcopy] = restype_ab.atoms[-1]
            atom.copied[forcopy].contents = dict(atom.contents.items())
            atom.copied[forcopy].name = atom.name + "R2"
            extra_b.append(atom.copied[forcopy])
            extra_b[-1].subsys = 2
        else:
            residue_type_b.atoms[i].copied[forcopy] = restype_ab.atoms[matchmap[i]]
    return extra_a, extra_b, rbmap


def _link_restypeb_atoms(residue_type_b, forcopy, matchmap):
    """

    :param residue_type_b:
    :param forcopy:
    :param matchmap:
    :return:
    """
    for atom in residue_type_b.atoms:
        for key, _ in atom.copied[forcopy].linked_atoms.items():
            for aton in atom.linked_atoms.get(key, []):
                if not (residue_type_b.atom2index(aton) in matchmap.keys()
                        and residue_type_b.atom2index(atom) in matchmap.keys()):
                    atom.copied[forcopy].Link_Atom(key, aton.copied[forcopy])


def _get_residue_ab(residue_type_a, residue_type_b, residue_a, forcopy, matchmap, match_a, match_b):
    """

    :param residue_type_a:
    :param residue_type_b:
    :param residue_a:
    :param forcopy:
    :param matchmap:
    :param match_a:
    :param match_b:
    :return:
    """
    restype_ab = residue_type_a.deepcopy(residue_type_a.name + "_" + residue_type_b.name, forcopy)

    extra_a, extra_b, rbmap = _get_extra_atoms_and_rbmap(restype_ab, residue_type_a, residue_type_b, residue_a, forcopy,
                                                         matchmap, match_a, match_b)

    for atomi in extra_a:
        for atomj in extra_b:
            atomi.Extra_Exclude_Atom(atomj)

    for atom, connect_set in residue_type_b.connectivity.items():
        for aton in connect_set:
            restype_ab.Add_Connectivity(atom.copied[forcopy], aton.copied[forcopy])

    for bond_entities in residue_type_b.bonded_forces.values():
        for bond_entity in bond_entities:
            tocopy = False
            for atom in bond_entity.atoms:
                if residue_type_b.atom2index(atom) not in matchmap.keys():
                    tocopy = True
                    break
            if tocopy:
                restype_ab.Add_Bonded_Force(bond_entity.deepcopy(forcopy))

    _link_restypeb_atoms(residue_type_b, forcopy, matchmap)

    for atom in residue_type_a.atoms:
        atom.copied.pop(forcopy)

    for atom in residue_type_b.atoms:
        atom.copied.pop(forcopy)
    return restype_ab, rbmap


def merge_dual_topology(mol, residue_a, residue_b, assign_a, assign_b,
                        tmcs=60, image_path=None, similarity_limit=0):
    """
    This **function** perturbs a residue in the molecule into another type in the dual topology way

    :param mol: the Molecule instance
    :param residue_a: the Residue which needs to be perturbed
    :param residue_b: the Residue which is perturbed to
    :param assign_a: the Assign instance corresponding to ``residue_a``
    :param assign_b: the Assign instance corresponding to ``residue_b``
    :param tmcs: the max time to find the max common structure
    :param image_path: the path to save the mcs image
    :param similarity_limit: the limitation of the similarity. The similarity is calculated by \
the tanimoto coefficient of the max common structure.
    :return: two molecules in the initial and final lambda stat respectively
    """
    build.Build_Bonded_Force(mol)
    build.Build_Bonded_Force(residue_b)

    from ...helper.rdkit import assign_to_rdmol, insert_atom_type_to_rdmol
    from rdkit.Chem import rdFMCS as MCS
    from rdkit.Chem import Draw, AllChem

    rdmol_a = assign_to_rdmol(assign_a, True)
    rdmol_b = assign_to_rdmol(assign_b, True)

    atom_type_dict = Xdict()
    insert_atom_type_to_rdmol(rdmol_a, residue_a, assign_a, atom_type_dict)
    insert_atom_type_to_rdmol(rdmol_b, residue_b, assign_b, atom_type_dict)
    print("FINDING MAXIMUM COMMON SUBSTRUCTURE\n")

    result = MCS.FindMCS([rdmol_a, rdmol_b], atomCompare=MCS.AtomCompare.CompareIsotopes, completeRingsOnly=True,
                         timeout=tmcs)
    rdmol_mcs = result.queryMol

    match_a = rdmol_a.GetSubstructMatch(rdmol_mcs)
    match_b = rdmol_b.GetSubstructMatch(rdmol_mcs)
    tanimoto = len(match_a)
    tanimoto /= len(assign_a.atoms) + len(assign_b.atoms) - tanimoto
    Xprint(f"similarity: {tanimoto}")
    assert tanimoto > similarity_limit, f"similarity (={tanimoto}) should be greater than {similarity_limit}"
    matchmap = {match_b[j]: match_a[j] for j in range(len(match_a))}
    if image_path:
        draw_a = assign_to_rdmol(assign_a, True)
        draw_b = assign_to_rdmol(assign_b, True)
        AllChem.Compute2DCoords(draw_a)
        AllChem.Compute2DCoords(draw_b)
        for atom in draw_a.GetAtoms():
            atom.SetProp("atomLabel", atom.GetSymbol())
        for atom in draw_b.GetAtoms():
            atom.SetProp("atomLabel", atom.GetSymbol())
        img = Draw.MolsToGridImage([draw_a, draw_b],
                                   molsPerRow=1,
                                   subImgSize=(1200, 600),
                                   highlightAtomLists=[match_a, match_b])
        img.save(image_path)
        f = Xopen(os.path.splitext(image_path)[0] + ".txt", "w")
        f.write(f"similarity: {tanimoto}\n")
        f.write(f"mcs atoms in r1: {match_a}\n\n")
        f.write(f"mcs atoms in r2: {match_b}\n\n")
        f.close()

    print("ALIGNING TOPOLOGY AND COORDINATE\n")

    residue_type_a = residue_a.type

    if isinstance(residue_b, Residue):
        residue_type_b = residue_b.type
    elif isinstance(residue_b, ResidueType):
        residue_type_b = residue_b
    else:
        raise TypeError

    _correct_residueb_coordinates(residue_a, residue_type_b, matchmap)

    forcopy = hash(str(time.time()))
    restype_ab, rbmap = _get_residue_ab(residue_type_a, residue_type_b, residue_a, forcopy, matchmap, match_a, match_b)

    restype_ba = restype_ab.deepcopy(residue_type_b.name + "_" + residue_type_a.name)

    build.Build_Bonded_Force(restype_ba)
    build.Build_Bonded_Force(restype_ab)

    nb14_extra_base.nb14_to_nb14_extra(restype_ba)
    nb14_extra_base.nb14_to_nb14_extra(restype_ab)
    for i, atomi in enumerate(restype_ab.atoms):
        if i < len(residue_type_a.atoms):
            atomi.contents.update(
                {key: value for key, value in residue_type_a.atoms[i].contents.items() if key != "name"})
        else:
            restype_ab.atoms[i].LJtype = "ZERO_LJ_ATOM"
            restype_ab.atoms[i].charge = 0
            restype_ab.atoms[i].subsys = 2
            restype_ba.atoms[i].subsys = 2

        if i in rbmap:
            restype_ba.atoms[i].contents.update(
                {key: value for key, value in residue_type_b.atoms[rbmap[i]].contents.items() if key != "name"})
        else:
            restype_ba.atoms[i].LJtype = "ZERO_LJ_ATOM"
            restype_ba.atoms[i].charge = 0
            restype_ab.atoms[i].subsys = 1
            restype_ba.atoms[i].subsys = 1

    mol_a = Molecule(mol.name + "A")
    mol_b = Molecule(mol.name + "B")

    for res in mol.residues:
        if res == residue_a:
            mol_a.Add_Residue(restype_ab)
            mol_b.Add_Residue(restype_ba)
        else:
            mol_a.Add_Residue(res)
            mol_b.Add_Residue(res)

    for reslink in mol.residue_links:
        mol_a.residue_links.add(reslink)
        mol_b.residue_links.add(reslink)

    build.Build_Bonded_Force(mol_a)
    build.Build_Bonded_Force(mol_b)

    return mol_a, mol_b


def merge_force_field(mol_a, mol_b, default_lambda, specific_lambda=None, intra_fep=False):
    """
    This **function** merges one molecule in two different force fields (two Molecule instances) into one

    :param mol_a: the Molecule instance in the initial lambda stat
    :param mol_b: the Molecule instance in the final lambda stat
    :param default_lambda: the lambda to scale the force if no ``specific_lambda`` is set for the force
    :param specific_lambda: a dict to map the force to its special scale factor lambda
    :param intra_fep: whether clear intramolecular non bonded interactions
    :return: the Molecule instance merged
    """
    if specific_lambda is None:
        specific_lambda = Xdict()
    build.Build_Bonded_Force(mol_a)
    build.Build_Bonded_Force(mol_b)

    nb14_extra_base.nb14_to_nb14_extra(mol_a)
    nb14_extra_base.nb14_to_nb14_extra(mol_b)

    assert len(mol_a.atoms) == len(mol_b.atoms)
    mol_a2mol_b = Xdict()
    for i, atom_a in enumerate(mol_a.atoms):
        mol_a2mol_b[atom_a] = mol_b.atoms[i]

    mol_r = mol_a.deepcopy()

    mol_r2mol_a = Xdict()
    mol_r2mol_b = Xdict()
    for i, atom_ret in enumerate(mol_r.atoms):
        mol_r2mol_a[atom_ret] = mol_a.atoms[i]
        mol_r2mol_b[atom_ret] = mol_b.atoms[i]
    mol_b2mol_r = {value: key for key, value in mol_r2mol_b.items()}
    mol_a2mol_r = {value: key for key, value in mol_r2mol_a.items()}

    charge_lambda = specific_lambda.get("charge", default_lambda)
    mass_lambda = specific_lambda.get("mass", default_lambda)

    for i, atom in enumerate(mol_r.atoms):
        atom.charge = mol_r2mol_a[atom].charge * (1 - charge_lambda) + mol_r2mol_b[atom].charge * charge_lambda
        atom.mass = mol_r2mol_a[atom].mass * (1 - mass_lambda) + mol_r2mol_b[atom].mass * mass_lambda
        atom.LJtypeB = mol_r2mol_b[atom].LJtype

    for forcename, rforces in mol_r.bonded_forces.items():
        if forcename in FEP_BONDED_FORCE_MERGE_RULE.keys():
            temp_lambda = specific_lambda.get(FEP_BONDED_FORCE_MERGE_RULE[forcename].get("lambda_name"),
                                              default_lambda)
            bforces = mol_b.bonded_forces.get(forcename, [])
            FEP_BONDED_FORCE_MERGE_RULE[forcename].get("merge_function")(mol_r, mol_a, mol_b,
                                                                         GlobalSetting.BondedForcesMap[forcename],
                                                                         rforces, bforces, temp_lambda, mol_r2mol_a,
                                                                         mol_a2mol_r, mol_r2mol_b, mol_b2mol_r)
        elif rforces:
            raise NotImplementedError(forcename + " is not supported for FEP to merge force field yet.")

    for forcename, parameters in FEP_BONDED_FORCE_MERGE_RULE.items():
        temp_lambda = specific_lambda.get(parameters["lambda_name"], default_lambda)

    return mol_r


set_global_alternative_names()



"""
This **module** is used to provide basic common math functions
"""
import numpy as np
from .namespace import  set_global_alternative_names


def get_rotate_matrix(r0, angle):
    """
    This **function** is used to get the rotation matrix by the axis and the angle

    :param r0: the axis to rotate around
    :param angle: the angle to rotate
    :return: the rotation matrix
    """
    cost = np.cos(angle)
    cost_one = 1 - cost
    sint = np.sin(angle)
    r0 /= np.linalg.norm(r0)
    return np.array([[r0[0] * r0[0] * cost_one + cost, r0[0] * r0[1] * cost_one - r0[2] * sint,
                      r0[0] * r0[2] * cost_one + r0[1] * sint],
                     [r0[0] * r0[1] * cost_one + r0[2] * sint, r0[1] * r0[1] * cost_one + cost,
                      r0[1] * r0[2] * cost_one - r0[0] * sint],
                     [r0[0] * r0[2] * cost_one - r0[1] * sint, r0[1] * r0[2] * cost_one + r0[0] * sint,
                      r0[2] * r0[2] * cost_one + cost]]).transpose()


def kabsch(positions1, positions2):
    """
    This **function** uses Kabsch algorithm to align two sets of positions

    :param positions1: the first set of positions, a numpy array with shape (N,3)
    :param positions2: the second set of positions, a numpy array with shape (N,3)
    :return: a tuple of 3 numpy arrays
             The first is the rotation matrix
             The second is the center of positions1.
             The third is the center of positions2.
    """
    positions1 = np.array(positions1, dtype=np.float32).reshape(-1, 3)
    positions2 = np.array(positions2, dtype=np.float32).reshape(-1, 3)
    center1 = np.mean(positions1, axis=0, keepdims=True)
    center2 = np.mean(positions2, axis=0, keepdims=True)
    x = positions1 - center1
    y = positions2 - center2
    r = np.einsum("kj,ki->ij", x, y)
    u, _, v = np.linalg.svd(r)
    return np.dot(u, v).transpose(), center1.reshape(-1), center2.reshape(-1)


def get_fibonacci_grid(n, origin, radius):
    """
    This **function** is used to get the sphere grid by fibonacci algorithm

    :param n: the number of grids
    :param origin: the origin of the sphere
    :param radius: the radius of the sphere
    :return: a numpy array of grid coordinates
    """
    n_ = np.arange(1, n + 1)
    factorn = (np.sqrt(5) - 1) * np.pi * n_
    out = np.zeros((n, 3))
    out[:, 2] = (2 * n_ - 1) / n - 1
    sqrtz = np.sqrt(1 - out[:, 2] * out[:, 2])
    out[:, 0] = sqrtz * np.cos(factorn)
    out[:, 1] = sqrtz * np.sin(factorn)
    out *= radius
    out += origin
    return out


def guess_element_from_mass(mass):
    """
    This **function** is used to guess element from its mass

    :param mass: an int or a float, the mass
    :return: the element name
    """
    elements = ["X", "H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne",
                "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc",
                "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge",
                "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc",
                "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe",
                "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
                "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os",
                "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr",
                "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf",
                "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt",
                "Ds", "Rg"]
    masses = [0.00000, 1.00794, 4.00260, 6.941, 9.012182, 10.811,
              12.0107, 14.0067, 15.9994, 18.9984032, 20.1797,
              22.989770, 24.3050, 26.981538, 28.0855, 30.973761,
              32.065, 35.453, 39.948, 39.0983, 40.078, 44.955910,
              47.867, 50.9415, 51.9961, 54.938049, 55.845, 58.9332,
              58.6934, 63.546, 65.409, 69.723, 72.64, 74.92160,
              78.96, 79.904, 83.798, 85.4678, 87.62, 88.90585,
              91.224, 92.90638, 95.94, 98.0, 101.07, 102.90550,
              106.42, 107.8682, 112.411, 114.818, 118.710, 121.760,
              127.60, 126.90447, 131.293, 132.90545, 137.327,
              138.9055, 140.116, 140.90765, 144.24, 145.0, 150.36,
              151.964, 157.25, 158.92534, 162.500, 164.93032,
              167.259, 168.93421, 173.04, 174.967, 178.49, 180.9479,
              183.84, 186.207, 190.23, 192.217, 195.078, 196.96655,
              200.59, 204.3833, 207.2, 208.98038, 209.0, 210.0, 222.0,
              223.0, 226.0, 227.0, 232.0381, 231.03588, 238.02891,
              237.0, 244.0, 243.0, 247.0, 247.0, 251.0, 252.0, 257.0,
              258.0, 259.0, 262.0, 261.0, 262.0, 266.0, 264.0, 269.0,
              268.0, 271.0, 272.0]
    if 3.8 > mass > 0.0:
        index = 1
    elif 208.99 > mass > 207.85:
        index = 83
    elif  58.8133 > mass > 56.50:
        index = 27
    else:
        index = 0
        for j in range(0, 111):
            if abs(mass - masses[j]) < 0.65:
                index = j
                break
    return elements[index]

set_global_alternative_names()



"""
This **module** is used to load and read
"""
import os
import warnings
import re

import numpy as np

from .helper import Molecule, Residue, ResidueType, AtomType, GlobalSetting, Xdict, Xopen, \
    set_real_global_variable, set_global_alternative_names


##########################################################################
# General Format
##########################################################################
def _mol2_atom(line, current_residue_index, current_residue, ignore_atom_type, temp, current_molecule,
               atom_residue_map):
    """

    :param line:
    :param current_residue_index:
    :param current_residue:
    :param ignore_atom_type:
    :param temp:
    :param current_molecule:
    :param atom_residue_map:
    :return:
    """
    words = line.split()
    if current_residue_index is None or int(words[6]) != current_residue_index:
        current_residue_index = int(words[6])
        if words[7] not in ResidueType.get_all_types():
            set_real_global_variable(words[7], ResidueType(name=words[7]))
            temp = True
        else:
            temp = False
        if current_residue:
            current_molecule.Add_Residue(current_residue)
        current_residue = Residue(ResidueType.get_type(words[7]))
    if ignore_atom_type:
        temp_atom_type = AtomType.get_type("UNKNOWN")
    else:
        temp_atom_type = AtomType.get_type(words[5])

    if temp:
        current_residue.type.Add_Atom(words[1], temp_atom_type, *words[2:5])
        current_residue.type.atoms[-1].Update(**{"charge[e]": float(words[8])})
    current_residue.Add_Atom(words[1], temp_atom_type, *words[2:5])
    current_residue.atoms[-1].Update(**{"charge[e]": float(words[8])})
    atom_residue_map[words[0]] = [words[1], current_residue, current_residue_index, temp, current_residue.atoms[-1],
                                  current_residue.type.atoms[-1]]
    return current_residue_index, current_residue, temp


def _mol2_bond(line, current_molecule, atom_residue_map):
    """

    :param line:
    :param current_molecule:
    :param atom_residue_map:
    :return:
    """
    words = line.split()
    if atom_residue_map[words[1]][1] == atom_residue_map[words[2]][1]:
        if atom_residue_map[words[1]][3]:
            atom_residue_map[words[1]][1].type.Add_Connectivity(atom_residue_map[words[1]][0],
                                                                atom_residue_map[words[2]][0])
        atom_residue_map[words[1]][1].Add_Connectivity(atom_residue_map[words[1]][0], atom_residue_map[words[2]][0])
    else:
        current_molecule.Add_Residue_Link(atom_residue_map[words[1]][4], atom_residue_map[words[2]][4])
        index_diff = atom_residue_map[words[1]][2] - atom_residue_map[words[2]][2]
        if abs(index_diff) == 1:
            if atom_residue_map[words[1]][3]:
                if index_diff < 0:
                    atom_residue_map[words[1]][1].type.tail = atom_residue_map[words[1]][0]
                else:
                    atom_residue_map[words[1]][1].type.head = atom_residue_map[words[1]][0]
            if atom_residue_map[words[2]][3]:
                if index_diff < 0:
                    atom_residue_map[words[2]][1].type.head = atom_residue_map[words[2]][0]
                else:
                    atom_residue_map[words[2]][1].type.tail = atom_residue_map[words[2]][0]


def _mol2_template_atom(line, current_residue_index, temp, current_residue, atom_residue_map):
    """

    :param line:
    :param current_residue_index:
    :param current_residue:
    :param ignore_atom_type:
    :param temp:
    :param current_molecule:
    :param atom_residue_map:
    :return:
    """
    words = line.split()
    if current_residue_index is None or int(words[6]) != current_residue_index:
        current_residue_index = int(words[6])
        if words[7] not in ResidueType.get_all_types():
            set_real_global_variable(words[7], ResidueType(name=words[7]))
            temp = True
        else:
            temp = False
        current_residue = ResidueType.get_type(words[7])

    temp_atom_type = AtomType.get_type(words[5])

    if temp:
        current_residue.Add_Atom(words[1], temp_atom_type, *words[2:5])
        current_residue.atoms[-1].Update(**{"charge[e]": float(words[8])})
    atom_residue_map[words[0]] = [words[1], current_residue, current_residue_index, temp]
    return current_residue_index, current_residue, temp


def _mol2_template_bond(line, atom_residue_map):
    """

    :param line:
    :param current_molecule:
    :param atom_residue_map:
    :return:
    """
    words = line.split()
    atom1_index, atom2_index = words[1], words[2]
    atom_name, residue_type, residue_index, is_new = 0, 1, 2, 3
    atom1_info, atom2_info = atom_residue_map[atom1_index], atom_residue_map[atom2_index]
    if atom1_info[residue_type] == atom2_info[residue_type] and atom1_info[is_new]:
        atom1_info[residue_type].Add_Connectivity(atom1_info[atom_name], atom2_info[atom_name])
    else:
        index_diff = atom1_info[residue_index] - atom2_info[residue_index]
        if abs(index_diff) == 1:
            if atom1_info[is_new]:
                if index_diff < 0:
                    atom1_info[residue_type].tail = atom1_info[atom_name]
                else:
                    atom1_info[residue_type].head = atom1_info[atom_name]
            if atom2_info[is_new]:
                if index_diff < 0:
                    atom2_info[residue_type].head = atom2_info[atom_name]
                else:
                    atom2_info[residue_type].tail = atom2_info[atom_name]


def _mol2_template(filename):
    """
    This is used to make loading more efficient when importing the force fields

    :param filename:
    :return:
    """
    current_residue_index = None
    current_residue = None
    atom_residue_map = Xdict()
    temp = None
    with open(filename) as f:
        flag = None
        for line in f:
            if not line.strip():
                continue
            if line.startswith("@<TRIPOS>"):
                flag = line[9:].strip()
            elif flag == "ATOM":
                current_residue_index, current_residue, temp = _mol2_template_atom(line, current_residue_index, temp,
                                                                                   current_residue, atom_residue_map)
            elif flag == "BOND":
                _mol2_template_bond(line, atom_residue_map)


def load_mol2(filename, ignore_atom_type=False, as_template=False):
    """
    This **function** is used to load a mol2 file

    :param filename: the name of the file to load
    :param ignore_atom_type: ignore the atom types in the mol2 file
    :param as_template: only read the mol2 file as some residue types and no molecule will created
        **New From 1.2.6.8**
    :return: a Molecule instance if as_template is False
    """
    if as_template:
        return _mol2_template(filename)
    with open(filename) as f:
        # 存储读的时候的临时信息，key是编号
        # value是list：原子名(0)、residue(1)、residue编号(2)、是否是新的residue type(3)、该原子(4)、residue type的最新原子(5)
        atom_residue_map = {}
        flag = None
        nline = 0
        current_molecule = None
        current_residue = None
        current_residue_index = None
        temp = None
        for line in f:
            if line.strip():
                nline += 1
            else:
                continue

            if line.startswith("@<TRIPOS>"):
                if flag == "ATOM":
                    current_molecule.Add_Residue(current_residue)
                flag = line[9:].strip()
                nline = 0
            elif flag == "MOLECULE":
                if nline == 1:
                    current_molecule = Molecule(line.strip())
            elif flag == "ATOM":
                current_residue_index, current_residue, temp = _mol2_atom(line, current_residue_index, current_residue,
                                                                          ignore_atom_type, temp, current_molecule,
                                                                          atom_residue_map)
            elif flag == "BOND":
                _mol2_bond(line, current_molecule, atom_residue_map)
    return current_molecule


def _pdb_ssbond_before(chain, residue_type_map, ssbonds):
    """

    :param chain:
    :param residue_type_map:
    :param ssbonds:
    :return:
    """
    for ssbond in ssbonds:
        res_a_index = chain[ssbond[15]][int(ssbond[17:21])]
        residue_type_map[res_a_index] = "CYX"
        res_b_index = chain[ssbond[29]][int(ssbond[31:35])]
        residue_type_map[res_b_index] = "CYX"


def _pdb_ssbond_after(chain, ssbonds, molecule):
    """

    :param chain:
    :param ssbonds:
    :param molecule:
    :return:
    """
    for ssbond in ssbonds:
        res_a_index = chain[ssbond[15]][int(ssbond[17:21])]
        res_b_index = chain[ssbond[29]][int(ssbond[31:35])]
        if res_a_index > res_b_index:
            res_a_index, res_b_index = res_b_index, res_a_index
        res_a = molecule.residues[res_a_index]
        res_b = molecule.residues[res_b_index]
        molecule.Add_Residue_Link(res_a.name2atom(res_a.type.connect_atoms["ssbond"]),
                                  res_b.name2atom(res_b.type.connect_atoms["ssbond"]))


def _pdb_link_after(chain, links, molecule):
    """

    :param chain:
    :param links:
    :param molecule:
    :return:
    """
    for link in links:
        chain_a, chain_b = link[21], link[51]
        res_a_pdb, res_b_pdb = int(link[22:26]), int(link[52:56])
        if chain_a not in chain or chain_b not in chain or \
                res_a_pdb not in chain[chain_a] or res_b_pdb not in chain[chain_a]:
            warnings.warn(f"The link between {chain_a}{res_a_pdb} and {chain_b}{res_b_pdb} is not valid")
            continue
        res_a_index = chain[link[21]][int(link[22:26])]
        res_b_index = chain[link[51]][int(link[52:56])]
        if res_a_index > res_b_index:
            res_a_index, res_b_index = res_b_index, res_a_index
        res_a = molecule.residues[res_a_index]
        res_b = molecule.residues[res_b_index]
        molecule.Add_Residue_Link(res_a.name2atom(link[12:16].strip()),
                                  res_b.name2atom(link[42:46].strip()))


def _pdb_add_atom(current_residue, atomname, x, y, z,
                  ignore_hydrogen, ignore_unknown_name, atom_map, atom_index):
    """
    add the atom to the residue
    """
    if ignore_hydrogen and (atomname.startswith("H") or
                            (len(atomname) > 1 and atomname[0] in "123" and atomname[1] == "H")):
        return False
    try:
        current_residue.Add_Atom(atomname, x=x, y=y, z=z)
        if atom_index not in atom_map:
            atom_map[atom_index] = current_residue.atoms[-1]
        return True
    except KeyError as ke:
        if ignore_unknown_name:
            return False
        raise ke


def _pdb_add_residue(filename, molecule, position_need, residue_type_map, ignore_hydrogen, ignore_unknown_name):
    """

    :param filename:
    :param molecule:
    :param position_need:
    :param residue_type_map:
    :param ignore_hydrogen:
    :param ignore_unknown_name:
    :return:
    """
    atom_map = Xdict()
    current_residue_count = -1
    current_residue_index = None
    current_insertion_code = None
    current_residue = None
    current_resname = None
    links = []
    with open(filename) as f:
        for line in f:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                extra = line[16]
                insertion_code = line[26]
                resindex = int(line[22:26])
                resname = line[17:20].strip()
                atomname = line[12:16].strip()
                atom_index = int(line[6:11])
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                if current_residue_index is None or current_residue_index != resindex or \
                        current_resname != resname or current_insertion_code != insertion_code:
                    current_residue_count += 1
                    if current_residue:
                        molecule.Add_Residue(current_residue)
                        if current_residue_index is not None and current_residue.type.tail and ResidueType.get_type(
                                residue_type_map[current_residue_count]).head:
                            links.append(len(molecule.residues))
                    current_residue = Residue(ResidueType.get_type(residue_type_map[current_residue_count]))
                    current_residue_index = resindex
                    current_insertion_code = insertion_code
                    current_resname = resname
                if extra not in (" ", position_need):
                    continue
                _pdb_add_atom(current_residue, atomname, x, y, z,
                              ignore_hydrogen, ignore_unknown_name, atom_map, atom_index)
            elif line.startswith("TER"):
                current_residue_index = None
                current_resname = None

    if current_residue:
        molecule.Add_Residue(current_residue)
    for count in links:
        res_a = molecule.residues[count]
        res_b = molecule.residues[count - 1]
        molecule.Add_Residue_Link(res_a.name2atom(res_a.type.head), res_b.name2atom(res_b.type.tail))
    return atom_map


def _pdb_judge_histone(judge_histone, residue_type_map, current_histone_information):
    """

    :param judge_histone:
    :param residue_type_map:
    :param current_histone_information:
    :return:
    """
    if judge_histone and residue_type_map and residue_type_map[-1] in GlobalSetting.HISMap["HIS"].keys():
        if current_histone_information["DeltaH"]:
            if current_histone_information["EpsilonH"]:
                residue_type_map[-1] = GlobalSetting.HISMap["HIS"][residue_type_map[-1]]["HIP"]
            else:
                residue_type_map[-1] = GlobalSetting.HISMap["HIS"][residue_type_map[-1]]["HID"]
        else:
            residue_type_map[-1] = GlobalSetting.HISMap["HIS"][residue_type_map[-1]]["HIE"]
        current_histone_information = {"DeltaH": False, "EpsilonH": False}


def _pdb_read_sequences(line, sequences):
    """

    :param line:
    :param sequences:
    :return:
    """
    chain_id = line[11]
    if chain_id not in sequences:
        sequences[chain_id] = []
    sequence = sequences[chain_id]
    sequence.extend(line[19:].split())


def _pdb_find_missing_residues(mol, sequences, chain, residue_type_map):
    """

    :param filename:
    :param sequences:
    :return:
    """
    if sequences:
        original_residue_names = [GlobalSetting.PDBResidueNameMap["save"].get(name, name) for name in residue_type_map]
        for chain_id, sequence in sequences.items():
            pdb_index_to_mol_index = chain[chain_id]
            name_index_to_mol_index = Xdict()
            pdb_index = list(pdb_index_to_mol_index.keys())
            pdb_index.sort()
            names_with_none = [None] * (pdb_index[-1] - pdb_index[0] + 1)
            for i in pdb_index:
                names_with_none[i - pdb_index[0]] = original_residue_names[pdb_index_to_mol_index[i]]
                name_index_to_mol_index[i - pdb_index[0]] = pdb_index_to_mol_index[i]
            offset = 0
            for offset in range(len(sequence) - len(names_with_none) + 1):
                if all(b in (a, None) for a, b in zip(sequence[offset:], names_with_none)):
                    break
            if offset:
                mol.set_missing_residues_info(None, name_index_to_mol_index[offset], sequence[:offset])
            tail_offset = len(sequence) - offset - len(names_with_none)
            if tail_offset:
                tail_index = name_index_to_mol_index[len(names_with_none) - 1]
                mol.set_missing_residues_info(tail_index, None, sequence[-tail_offset:])
            find_none_index = None
            for i, name in enumerate(names_with_none):
                if find_none_index is None and name is None:
                    find_none_index = i
                elif find_none_index is not None and name is not None:
                    start = name_index_to_mol_index[find_none_index - 1]
                    end = name_index_to_mol_index[i]
                    mol.set_missing_residues_info(start, end, sequence[find_none_index:i])
                    find_none_index = None


def _pdb_connects(mol, connects, atom_map):
    """

    :param mol:
    :param connects:
    :param atom_map:
    :return:
    """
    for connect in connects:
        atom = int(connect[6:11])
        if atom not in atom_map:
            continue
        atom_ = atom_map[atom]
        connect = connect[11:]
        for i in range(0, len(connect), 5):
            to_connect = connect[i:i+5]
            if not to_connect.strip():
                break
            to_connect = int(to_connect)
            if to_connect not in atom_map:
                continue
            to_connect_ = atom_map[to_connect]
            if atom > to_connect or atom_.residue == to_connect_.residue or mol.get_residue_link(atom_, to_connect_):
                continue
            mol.add_residue_link(atom_, to_connect_)
            print(atom_, atom, to_connect_, to_connect)


def load_pdb(filename, judge_histone=True, position_need="A", ignore_hydrogen=False,
             ignore_unknown_name=False, ignore_seqres=True, ignore_conect=True):
    """
    This **function** is used to load a pdb file

    :param filename: the name of the file to load
    :param judge_histone: judge the protonized state of the histone residues
    :param position_need: the position character to read
    :param ignore_hydrogen: do not read the atom with a name beginning with "H" or "[123]H"
    :param ignore_unknown_name: do not read the atom with an unknown name **New From 1.2.6.4**
    :param ignore_seqres: do not read the SEQRES lines **New From 1.2.6.7**
    :param ignore_conect: do not read the CONECT lines **New From 1.2.6.7**
    :return: a Molecule instance
    """
    molecule = Molecule(os.path.splitext(os.path.basename(filename))[0])
    chain = Xdict(not_found_method=lambda key: Xdict())
    sequences = Xdict()
    ssbonds = []
    links = []
    connects = []
    residue_type_map = []
    insertion_count = 0
    current_residue_count = -1
    current_insertion_code = None
    current_residue_index = None
    current_resname = None
    current_histone_information = {"DeltaH": False, "EpsilonH": False}
    with open(filename) as f:
        for line in f:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                resindex = int(line[22:26]) + insertion_count
                insertion_code = line[26]
                chain_id = line[21]
                if not chain_id.strip():
                    chain_id = 0
                resname = line[17:20].strip()
                atomname = line[12:16].strip()
                if current_residue_index is None:
                    _pdb_judge_histone(judge_histone, residue_type_map, current_histone_information)
                    current_residue_count += 1
                    current_resname = resname
                    current_insertion_code = insertion_code
                    if resname in GlobalSetting.PDBResidueNameMap["head"].keys():
                        resname = GlobalSetting.PDBResidueNameMap["head"][resname]
                    residue_type_map.append(resname)
                    current_residue_index = resindex
                    chain[chain_id][resindex] = current_residue_count
                elif (current_residue_index != resindex or current_insertion_code != insertion_code) or \
                        current_resname != resname:
                    _pdb_judge_histone(judge_histone, residue_type_map, current_histone_information)
                    current_residue_count += 1
                    current_resname = resname
                    current_insertion_code = insertion_code
                    if insertion_code != " ":
                        insertion_count += 1
                        resindex += 1
                    current_residue_index = resindex
                    chain[chain_id][resindex] = current_residue_count
                    residue_type_map.append(resname)
                if judge_histone and resname in GlobalSetting.HISMap["HIS"].keys():
                    if atomname == GlobalSetting.HISMap["DeltaH"]:
                        current_histone_information["DeltaH"] = True
                    elif atomname == GlobalSetting.HISMap["EpsilonH"]:
                        current_histone_information["EpsilonH"] = True

            elif line.startswith("TER"):
                current_residue_index = None
                current_resname = None
                current_insertion_code = None
                insertion_count = 0
                if residue_type_map[-1] in GlobalSetting.PDBResidueNameMap["tail"].keys():
                    residue_type_map[-1] = GlobalSetting.PDBResidueNameMap["tail"][residue_type_map[-1]]
                _pdb_judge_histone(judge_histone, residue_type_map, current_histone_information)

            elif line.startswith("SSBOND"):
                ssbonds.append(line)

            elif line.startswith("LINK"):
                links.append(line)

            elif line.startswith("SEQRES") and not ignore_seqres:
                _pdb_read_sequences(line, sequences)

            elif line.startswith("CONECT") and not ignore_conect:
                connects.append(line)

    current_residue_index = None
    if residue_type_map[-1] in GlobalSetting.PDBResidueNameMap["tail"].keys():
        residue_type_map[-1] = GlobalSetting.PDBResidueNameMap["tail"][residue_type_map[-1]]

    _pdb_ssbond_before(chain, residue_type_map, ssbonds)
    atom_map = _pdb_add_residue(filename, molecule, position_need,
                                residue_type_map, ignore_hydrogen, ignore_unknown_name)
    _pdb_ssbond_after(chain, ssbonds, molecule)
    _pdb_link_after(chain, links, molecule)
    _pdb_find_missing_residues(molecule, sequences, chain, residue_type_map)
    _pdb_connects(molecule, connects, atom_map)

    return molecule


##########################################################################
# SPONGE Format
##########################################################################
def load_coordinate(filename, mol=None):
    """
    This **function** is used to read the SPONGE coordinate in file

    :param filename: the coordinate file to load
    :param mol: the molecule or residue to load the coordinate into
    :return: two numpy arrays, representing the coordinates and the box information respectively
    """
    with Xopen(filename, "r") as f:
        atom_numbers = int(f.readline().split()[0])
        crd = np.zeros((atom_numbers, 3), dtype=np.float32)
        box = np.zeros(6, dtype=np.float32)
        for i in range(atom_numbers):
            crd[i][:] = np.array([float(x) for x in f.readline().split()])
        box = np.array(f.readline().split(), dtype=np.float32)
    if mol:
        for i, atom in enumerate(mol.atoms):
            atom.x = crd[i][0]
            atom.y = crd[i][1]
            atom.z = crd[i][2]
        if isinstance(mol, Molecule):
            mol.box_length = box
    return crd, box


##########################################################################
# amber Format
##########################################################################
def _frcmod_nb14(line, atoms):
    nb14ee = re.findall(r"SCEE=[\d+\.]+", line)
    nb14lj = re.findall(r"SCNB=[\d+\.]+", line)
    if not nb14ee and not nb14lj:
        return ""
    nb14ee = 1.0 / float(nb14ee[0].split("=")[1]) if nb14ee else 1.0 / 1.2
    nb14lj = 1.0 / float(nb14lj[0].split("=")[1]) if nb14lj else 1.0 / 2.0
    return f"{atoms[0]}-{atoms[3]} {nb14ee} {nb14lj}\n"


def _frcmod_cmap(line, cmap, temp_cmp, cmap_flag):
    """

    :param line:
    :param cmap:
    :param temp_cmp:
    :param cmap_flag:
    :return:
    """
    if line.startswith("%FLAG"):
        if "CMAP_COUNT" in line:
            if temp_cmp:
                for res in temp_cmp["residues"]:
                    cmap[f"C-N-{res}@XC-C-N"] = {"resolution": temp_cmp["info"]["resolution"],
                                                 "parameters": temp_cmp["info"]["parameters"]}
            temp_cmp = {"residues": [], "info": {"resolution": 24, "count": int(line.split()[-1]), "parameters": []}}
            cmap_flag = "CMAP_COUNT"
        elif "CMAP_RESOLUTION" in line:
            temp_cmp["info"]["resolution"] = int(line.split()[-1])
            cmap_flag = "CMAP_RESOLUTION"
        elif "CMAP_RESLIST" in line:
            cmap_flag = "CMAP_RESLIST"
        elif "CMAP_TITLE" in line:
            cmap_flag = "CMAP_TITLE"
        elif "CMAP_PARAMETER" in line:
            cmap_flag = "CMAP_PARAMETER"
    elif cmap_flag == "CMAP_RESLIST":
        temp_cmp["residues"].extend(line.split())
    elif cmap_flag == "CMAP_PARAMETER":
        temp_cmp["info"]["parameters"].extend([float(x) for x in line.split()])
    return temp_cmp, cmap_flag


def _frcmod_atoms_words(line, n, last_atoms=None):
    """

    :param line:
    :param n:
    :return:
    """
    if line[0] != " ":
        return [word.strip() for word in line[:n].split("-")], line[n:].split()
    return [last_atoms, line[n:].split()]


def load_frcmod(filename, nbtype="RE"):
    """
    This **function** is used to load a frcmod file

    :param filename: the name of the file to load
    :param nbtype: the non-bonded interaction recording type in the frcmod file.
    :return: a list of strings, including atoms, bonds, angles, propers, impropers, ljs, cmap information respectively
    """
    with open(filename) as f:
        f.readline()
        flag = None
        atom_types = {}  # 元素符号和质量
        bonds = "name  k[kcal/mol·A^-2]    b[A]\n"
        angles = "name  k[kcal/mol·rad^-2]    b[degree]\n"
        propers = "name  k[kcal/mol]    phi0[degree]    periodicity    reset\n"
        reset = 1
        impropers = "name  k[kcal/mol]    phi0[degree]    periodicity\n"
        cmap = {}
        cmap_flag = None
        temp_cmp = {"residues": []}
        if nbtype == "SK":
            raise NotImplementedError
        if nbtype == "AC":
            ljs = "name A[kcal/mol·A^-12]   B[kcal/mol·A^-6]\n"
        elif nbtype == "RE":
            ljs = "name rmin[A]   epsilon[kcal/mol]\n"

        for line in f:
            if not line.strip():
                continue
            words = line.split()
            if flag != "CMAP" and len(words) == 1:
                flag = line.strip()
            elif flag[:4] == "MASS":
                atom_types[words[0]] = words[1]
            elif flag[:4] == "BOND":
                atoms, words = _frcmod_atoms_words(line, 5)
                bonds += "-".join(atoms) + "\t" + words[0] + "\t" + words[1] + "\n"
            elif flag[:4] == "ANGL":
                atoms, words = _frcmod_atoms_words(line, 8)
                angles += "-".join(atoms) + "\t" + words[0] + "\t" + words[1] + "\n"
            elif flag[:4] == "DIHE":
                atoms, words = _frcmod_atoms_words(line, 11)
                propers += "-".join(atoms) + "\t" + str(float(words[1]) / int(words[0])) + "\t" + words[2] + "\t" + str(
                    abs(int(float(words[3])))) + "\t" + str(reset) + "\n"
                if int(float(words[3])) < 0:
                    reset = 0
                else:
                    reset = 1
            elif flag[:4] == "IMPR":
                atoms, words = _frcmod_atoms_words(line, 11)
                impropers += "-".join(atoms) + "\t" + words[0] + "\t" + words[1] + "\t" + str(
                    int(float(words[2]))) + "\n"
            elif flag[:4] == "NONB":
                words = line.split()
                ljs += words[0] + "-" + words[0] + "\t" + words[1] + "\t" + words[2] + "\n"
            elif flag[:4] == "CMAP":
                temp_cmp, cmap_flag = _frcmod_cmap(line, cmap, temp_cmp, cmap_flag)

    for res in temp_cmp["residues"]:
        cmap[f"C-N-{res}@XC-C-N"] = {"resolution": temp_cmp["info"]["resolution"],
                                     "parameters": temp_cmp["info"]["parameters"]}
    atoms = "name  mass  LJtype\n"
    for atom, mass in atom_types.items():
        atoms += atom + "\t" + mass + "\t" + atom + "\n"
    toret = [atoms, bonds, angles, propers, impropers, ljs, cmap]
    return toret


def _parmdat_read_harmonic_bonds(f, bonds, n):
    """

    :param f:
    :param bonds:
    :param n:
    :return:
    """
    for line in f:
        if not line.strip():
            break
        atoms, words = _frcmod_atoms_words(line, n)
        bonds += "-".join(atoms) + "\t" + words[0] + "\t" + words[1] + "\n"
    return bonds


def load_parmdat(filename):
    """
    This **function** is used to load a parmdat file

    :param filename: the name of the file to load
    :return: a list of strings, including atoms, bonds, angles, propers, impropers, ljs, nb14s information respectively
    """
    with open(filename) as f:
        f.readline()
        # 读原子
        atom_types = {}  # 元素符号和质量
        lj_types = Xdict()  # 元素符号和LJ类型
        for line in f:
            if not line.strip():
                break

            words = line.split()
            atom_types[words[0]] = words[1]
            lj_types[words[0]] = words[0]
        f.readline()
        # 读键长
        bonds = "name  k[kcal/mol·A^-2]    b[A]\n"
        bonds = _parmdat_read_harmonic_bonds(f, bonds, 5)

        # 读键角
        angles = "name  k[kcal/mol·rad^-2]    b[degree]\n"
        angles = _parmdat_read_harmonic_bonds(f, angles, 8)

        # 读恰当二面角
        reset = 1
        propers = "name  k[kcal/mol]    phi0[degree]    periodicity    reset\n"
        nb14s = "name    kLJ     kee\n"
        atoms = None
        for line in f:
            if not line.strip():
                break
            last_atoms = atoms
            atoms, words = _frcmod_atoms_words(line, 11, last_atoms)
            nb14s += _frcmod_nb14(line, atoms)
            propers += "-".join(atoms) + "\t" + str(float(words[1]) / int(words[0])) + "\t" + words[2] + "\t" + str(
                abs(int(float(words[3])))) + "\t" + str(reset) + "\n"
            if int(float(words[3])) < 0:
                reset = 0
            else:
                reset = 1

        # 读非恰当二面角
        impropers = "name  k[kcal/mol]    phi0[degree]    periodicity\n"
        for line in f:
            if not line.strip():
                break
            atoms, words = _frcmod_atoms_words(line, 11)
            impropers += "-".join(atoms) + "\t" + words[0] + "\t" + words[1] + "\t" + str(
                int(float(words[2]))) + "\n"

        # 跳过水的信息
        f.readline()
        f.readline()

        # 读LJ种类
        for line in f:
            if not line.strip():
                break
            atoms = line.split()
            atom0 = atoms.pop(0)
            for atom in atoms:
                lj_types[atom] = atom0

        # 读LJ信息
        word = f.readline().split()[1]
        if word == "SK":
            raise NotImplementedError
        if word == "AC":
            ljs = "name A[kcal/mol·A^-12]   B[kcal/mol·A^-6]\n"
        elif word == "RE":
            ljs = "name rmin[A]   epsilon[kcal/mol]\n"

        for line in f:
            if not line.strip():
                break
            words = line.split()
            ljs += words[0] + "-" + words[0] + "\t" + words[1] + "\t" + words[2] + "\n"

    atoms = "name  mass  LJtype\n"
    for atom, mass in atom_types.items():
        atoms += atom + "\t" + mass + "\t" + lj_types[atom] + "\n"
    toret = [atoms, bonds, angles, propers, impropers, ljs, nb14s]
    return toret


def load_rst7(filename, mol=None):
    """
    This **function** is used to load a rst7 file

    :param filename: the name of the file to load
    :param mol: the molecule to load the coordinates
    :return: a tuple including coordinates and box information
    """
    crds = []
    with open(filename) as f:
        f.readline()
        words = f.readline().split()
        atom_numbers = int(words[0])
        line = ""
        for line in f:
            words = line.split()
            while words and len(crds) < atom_numbers * 3:
                crds.append(float(words.pop(0)))
        box = [float(i) for i in line.split()]
    crds = np.array(crds).reshape((-1, 3))
    mol.box_length = box[:3]
    count = -1
    for residue in mol.residues:
        for atom in residue.atoms:
            count += 1
            atom.x = crds[count][0]
            atom.y = crds[count][1]
            atom.z = crds[count][2]

    return crds, box


##########################################################################
# GROMACS Format
##########################################################################
class GromacsTopologyIterator():
    """
    This **class** is used to read a GROMACS topology file

    usage example::

        f = GromacsTopologyIterator("example.itp")
        for line in f:
            print(line)

    :param filename: the name of the file to read
    :param macros: the macros used to read the Gromacs topology file

    """

    def __init__(self, filename=None, macros=None):
        self.files = []
        self.filenames = []

        self.flag = ""
        self.macro_define_stat = []

        if macros:
            self.defined_macros = macros
        else:
            self.defined_macros = {}
        if filename:
            self._add_iterator_file(filename)

    def __iter__(self):
        self.flag = ""
        self.macro_define_stat = []
        return self

    def __next__(self):
        while self.files:
            f = self.files[-1]
            line = f.readline()
            if line:
                line = self._line_preprocess(line)
                if line[0] == "#":
                    line = self._line_define(line)
                elif self.macro_define_stat and not self.macro_define_stat[-1]:
                    line = next(self)
                elif "[" in line and "]" in line:
                    self.flag = line[1:-1].strip()
                    line = next(self)
                for macro, tobecome in self.defined_macros.items():
                    line = line.replace(macro, tobecome)
                return line

            f.close()
            self.files.pop()
            self.filenames.pop()

        raise StopIteration

    def _add_iterator_file(self, filename):
        """

        :param filename:
        :return:
        """
        if self.files:
            filename = os.path.abspath(os.path.join(os.path.dirname(self.filenames[-1]), filename.replace('"', '')))
        else:
            filename = os.path.abspath(filename.replace('"', ''))

        f = Xopen(filename, "r")
        self.files.append(f)
        self.filenames.append(filename)

    def _line_preprocess(self, line):
        """

        :param line:
        :return:
        """
        line = line.strip()
        comment = line.find(";")
        if comment >= 0:
            line = line[:comment]
        while line and line[-1] == "\\":
            line = line[:-1] + " " + next(self).strip()
        if not line:
            line = next(self)
        return line

    def _line_define(self, line):
        """

        :param line:
        :return:
        """
        words = line.split()
        if words[0] == "#ifdef":
            macro = words[1]
            if self.macro_define_stat and not self.macro_define_stat[-1]:
                self.macro_define_stat.append(False)
            elif macro in self.defined_macros.keys():
                self.macro_define_stat.append(True)
            else:
                self.macro_define_stat.append(False)
        elif words[0] == "#else":
            if len(self.macro_define_stat) <= 1 or self.macro_define_stat[-2]:
                self.macro_define_stat[-1] = not self.macro_define_stat[-1]
        elif words[0] == "#endif":
            self.macro_define_stat.pop()
        elif self.macro_define_stat and not self.macro_define_stat[-1]:
            next(self)
        elif words[0] == "#define":
            if len(words) > 2:
                self.defined_macros[words[1]] = line[line.find(words[1]) + len(words[1]):].strip()
            else:
                self.defined_macros[words[1]] = ""
        elif words[0] == "#include":
            self._add_iterator_file(words[1])
        elif words[0] == "#undef":
            self.defined_macros.pop(words[1])
        elif words[0] == "#error":
            raise Exception(line)
        line = next(self)
        return line


def _ffitp_dihedrals(line, output):
    """

    :param line:
    :param output:
    :return:
    """
    words = line.split()
    func = words[4]
    if func == "1":
        output["dihedrals"] += "-".join(words[:4]) + " " + " ".join(words[5:]) + " 0\n"
    elif func == "2":
        temp1 = [words[1], words[2], words[0], words[3]]
        temp2 = [words[1], words[2], words[3], words[0]]
        if words[0][0].upper() in ("C", "N", "S"):
            output["impropers"] += "-".join(temp1) + " {b} {k}".format(b=float(words[5]), k=float(words[6]) / 2) + "\n"
        else:
            output["impropers"] += "-".join(temp2) + " {b} {k}".format(b=float(words[5]), k=float(words[6]) / 2) + "\n"
    elif func == "9":
        for i in range(5, len(words), 20):
            output["dihedrals"] += "-".join(words[:4]) + " " + " ".join(words[i:i + 3]) + " 0\n"
    else:
        raise NotImplementedError


def load_ffitp(filename, macros=None):
    """
    This **function** is used to load a fftip file

    .. ATTENTION::

        This is used to read a force field itp (ffitp) file instead of a simple itp file for a molecule.

    :param filename: the name of the file to load
    :param macros: the macros used to read the Gromacs topology file
    :return: a dict, which stores the name of the forcefield term - the corresponding information mapping
    """
    iterator = GromacsTopologyIterator(filename, macros)
    output = Xdict()
    output["nb14"] = "name  kLJ  kee\n"
    output["atomtypes"] = "name mass charge[e] LJtype\n"
    output["bonds"] = "name b[nm] k[kJ/mol·nm^-2]\n"
    output["angles"] = "name b[degree] k[kJ/mol·rad^-2]\n"
    output["Urey-Bradley"] = "name b[degree] k[kJ/mol·rad^-2] r13[nm] kUB[kJ/mol·nm^-2]\n"
    output["dihedrals"] = "name phi0[degree] k[kJ/mol] periodicity  reset\n"
    output["impropers"] = "name phi0[degree] k[kJ/mol·rad^-2]\n"
    output["cmaps"] = Xdict()
    for line in iterator:
        if iterator.flag == "":
            continue
        if iterator.flag == "defaults":
            words = line.split()
            assert int(words[0]) == 1, "SPONGE Only supports Lennard-Jones now"
            if int(words[1]) == 1:
                output["LJ"] = "name A[kJ/mol·nm^6] B[kJ/mol·nm^12]\n"
                output["nb14_extra"] = "name A[kJ/mol·nm^6] B[kJ/mol·nm^12] kee\n"
            else:
                output["LJ"] = "name sigma[nm] epsilon[kJ/mol] \n"
                output["nb14_extra"] = "name sigma[nm] epsilon[kJ/mol] kee\n"
            fudge_lj = float(words[3])
            fudge_qq = float(words[4])
            if words[2] == "yes":
                output["nb14"] += "X-X {fudgeLJ} {fudgeQQ}\n".format(fudgeLJ=fudge_lj, fudgeQQ=fudge_qq)

        elif iterator.flag == "atomtypes":
            words = line.split()
            output["atomtypes"] += "{type} {mass} {charge} {type}\n".format(type=words[0], mass=float(words[2]),
                                                                            charge=float(words[3]))
            output["LJ"] += "{type}-{type} {V} {W}\n".format(type=words[0], V=float(words[5]), W=float(words[6]))
        elif iterator.flag == "pairtypes":
            words = line.split()
            if len(words) <= 3:
                output["nb14"] += "{atom1}-{atom2} {kLJ} {kee}\n".format(atom1=words[0], atom2=words[1], kLJ=fudge_lj,
                                                                         kee=fudge_qq)
            elif words[2] == "1":
                output["nb14_extra"] += "{atom1}-{atom2} {V} {W} {kee}\n".format(atom1=words[0], atom2=words[1],
                                                                                 V=float(words[3]), W=float(words[4]),
                                                                                 kee=fudge_qq)
                output["nb14"] += "{atom1}-{atom2} 0 0\n".format(atom1=words[0], atom2=words[1])
            elif words[2] == "2":
                raise NotImplementedError
        elif iterator.flag == "bondtypes":
            words = line.split()
            func = words[2]
            if func == "1":
                output["bonds"] += "{atom1}-{atom2} {b} {k}\n".format(atom1=words[0], atom2=words[1], b=float(words[3]),
                                                                      k=float(words[4]) / 2)
            else:
                raise NotImplementedError
        elif iterator.flag == "angletypes":
            words = line.split()
            func = words[3]
            if func == "1":
                output["angles"] += "-".join(words[:3]) + " {b} {k}".format(b=float(words[4]),
                                                                            k=float(words[5]) / 2) + "\n"
            elif func == "5":
                output["Urey-Bradley"] += "-".join(words[:3]) + " {b} {k} {b2} {k2}".format(b=float(words[4]),
                                                                                            k=float(words[5]) / 2,
                                                                                            b2=float(words[6]),
                                                                                            k2=float(
                                                                                                words[7]) / 2) + "\n"
            else:
                raise NotImplementedError
        elif iterator.flag == "dihedraltypes":
            _ffitp_dihedrals(line, output)
        elif iterator.flag == "cmaptypes":
            words = line.split()
            output["cmaps"]["-".join(words[:5])] = {"resolution": int(words[7]),
                                                    "parameters": list(map(lambda x: float(x) / 4.184, words[8:]))}
        elif iterator.flag == "nonbond_params":
            words = line.split()
            output["LJ"] += "{type1}-{type2} {V} {W}\n".format(type1=words[0], type2=words[1], V=float(words[3]),
                                                               W=float(words[4]))
    return output

set_global_alternative_names()



"""
This **package** sets the basic configuration of RSFF2C
"""
from ...helper import source, Xdict
source("..ff14sb")
source("...base.cmap_base")

def _init():
    """
    Initialize the module
    """
    rsff2c_dihedral = Generate_New_Bonded_Force_Type("dihedral", "1-2-3-4",
                                                     {"k": float, "phi0": float, "periodicity": int},
                                                     False, ["k", "phi0", "periodicity"])
    rsff2c_dihedral.Set_Property_Unit("k", "energy", "kcal/mol")
    rsff2c_dihedral.Set_Property_Unit("phi0", "angle", "rad")
    @rsff2c_dihedral.Type_Name_Getter
    def build_check(atoms):
        return "-".join([atom.residue.name + "@" + atom.name if len(atom.residue.name) <= 3
                         else atom.residue.name[-3:] + "@" + atom.name for atom in atoms])
    source_data = """
    dih_E_chi2 = '-0.457   0.453  -0.105   0.269'
    dih_E_chi3 = ' 0.028   0.201   0.052   0.025'
    dih_Q_chi2 = ' 0.603   0.672   0.341   0.311'
    dih_Q_chi3 = '-0.187   0.012  -0.440   0.032'
    dih_N_chi2 = '-0.343   0.484  -0.154  -0.108'
    """.strip().replace("\n", "=").split("=")
    coeffs = Xdict({source_data[i].strip() : [float(j) for j in source_data[i + 1].split("'")[1].split()]
                    for i in range(0, len(source_data), 2)})
    phase = Xdict({key: [ 3.1415926 if x > 0 else 0 for x in coeff] for key, coeff in coeffs.items()})
    to_update = "name k phi0 periodicity reset\n"
    names = Xdict({"GLU@CA-GLU@CB-GLU@CG-GLU@CD": "dih_E_chi2",
                   "GLU@CB-GLU@CG-GLU@CD-GLU@OE1": "dih_E_chi3",
                   "GLU@CB-GLU@CG-GLU@CD-GLU@OE2": "dih_E_chi3",
                   "GLN@CA-GLN@CB-GLN@CG-GLN@CD": "dih_Q_chi2",
                   "GLN@CB-GLN@CG-GLN@CD-GLN@OE1": "dih_Q_chi3",
                   "ASN@CA-ASN@CB-ASN@CG-ASN@OD1": "dih_N_chi2"})
    for name, pname in names.items():
        to_update += f"{name} {abs(coeffs[pname][0])} {phase[pname][0]} 1 1\n"
        to_update += f"{name} {abs(coeffs[pname][1])} {phase[pname][1]} 2 0\n"
        to_update += f"{name} {abs(coeffs[pname][2])} {phase[pname][2]} 3 0\n"
        to_update += f"{name} {abs(coeffs[pname][3])} {phase[pname][3]} 4 0\n"

    rsff2c_dihedral.New_From_String(to_update)

    rsff2c_cmap = Generate_New_Bonded_Force_Type("cmap",
                                                 "1-2-3-4-5",
                                                 {"resolution": int, "parameters": list}, False)
    @rsff2c_cmap.Type_Name_Getter
    def _(atoms):
        atom_names = ["G" if ((atom.residue.name == "VAL" and atom.name == "CG2") or
                              (atom.residue.name != "VAL" and atom.name in ('CG', 'OG', 'SG', 'CG1', 'OG1')))
                      else atom.name for atom in atoms]
        res_name = atoms[2].residue.name
        if "-".join(atom_names) == "C-N-CA-C-N" and res_name != "GLY":
            return "C-N-ALA@CA-C-N"
        atom_names[2] = res_name + "@" + atom_names[2]
        return "-".join(atom_names)

    with open(os.path.join(AMBER_DATA_DIR, "RSFF2C.dat")) as f:
        source_data = f.read().strip().replace("%FLAG", "%FORMAT(8(F9.5))").split("%FORMAT(8(F9.5))")[1:]
    names = [line.split("COMMENT")[1].split() for line in source_data[::2]]
    temp_map = Xdict({"phi/psi": "C-N-%s@CA-C-N",
                      "chi/phi": "G-CB-%s@CA-N-C",
                      "chi/psi": "G-CB-%s@CA-C-N"})
    names = [temp_map[name[1]]%name[0] for name in names]
    parameters = [[float(word) for word in line.split()] for line in source_data[1::2]]
    to_update = Xdict()
    for name, parameter in zip(names, parameters):
        to_update[name] = {"resolution":24, "parameters": parameter}
    rsff2c_cmap.New_From_Dict(to_update)


_init()

Xprint("""Reference for RSFF2C:
  Wei Kang, Fan Jiang, Yun-Dong Wu
    Universal Implementation of a Residue-Specific Force Field Based on CMAP Potentials and Free Energy Decomposition
    Journal of Chemical Theory and Computation 2018, 14(8), 4474–4486
    DOI: 10.1021/acs.jctc.8b00285
""")



"""
This **module** implements the terminal commands
"""
import os
import sys
import unittest
import multiprocessing as mpc

from ..helper import source, GlobalSetting, Xopen, Xprint
from ..mdrun import run

class TestMyPackage(unittest.TestCase):
    """
    This **class** does the unit tests for Xponge

    :param methodName: the method name of the unit test
    :param args: arguments from argparse
    :return: None
    """
    def __init__(self, methodName='runTest', args=None):
        super().__init__(methodName)
        self.args = args

    @classmethod
    def get_test_suite(cls, name, args):
        suite = unittest.TestSuite()
        test_name = "test_" + name
        suite.addTest(cls(test_name, args))
        return suite

    def test_base(self):
        """
        This **function** does the basic test for Xponge

        :param args: arguments from argparse
        :return: None
        """
        args = self.args
        source("..")
        source("..forcefield.amber.ff14sb")
        source("..forcefield.amber.tip3p")
        source("__main__")

        t = ACE + ALA * 10 + NME

        for i in range(1, len(t.residues) - 1):
            head = t.residues[i - 1]
            res = t.residues[i]
            tail = t.residues[i + 1]
            Impose_Dihedral(t, head.C, res.N, res.CA, res.C, -3.1415926 / 3)
            Impose_Dihedral(t, res.N, res.CA, res.C, tail.N, -3.1415926 / 3)

        Save_Mol2(t, f"{args.o}.mol2")
        c = int(round(t.charge))
        Add_Solvent_Box(t, WAT, 10)
        Solvent_Replace(t, lambda res: res.type.name == "WAT", {CL: 30 + c, K: 30})
        t.residues.sort(key=lambda residue: {"CL": 2, "K": 1, "WAT": 3}.get(residue.type.name, 0))
        Save_PDB(t, f"{args.o}.pdb")
        Save_SPONGE_Input(t, f"{args.o}")

        f = Xopen(f"{args.o}_LJ.txt", "r")
        atom_number, lj_number = [int(i) for i in f.readline().split()]
        self.assertEqual(atom_number, 2923, "LJ_in_file wrong")
        self.assertEqual(lj_number, 11, "LJ_in_file wrong")
        f.close()

        f = Xopen(f"{args.o}_dihedral.txt", "r")
        dihedral_number, = [int(i) for i in f.readline().split()]
        self.assertEqual(dihedral_number, 303, "dihedral_in_file wrong")
        f.close()

        f = Xopen(f"{args.o}_exclude.txt", "r")
        atom_number, exclude_number = [int(i) for i in f.readline().split()]
        self.assertEqual(atom_number, 2923, "exclude_in_file wrong")
        self.assertEqual(exclude_number, 3326, "exclude_in_file wrong")
        f.close()

    def test_charmm27(self):
        """
        This **function** does the CHARMM27 test for Xponge

        :param args: arguments from argparse
        :return: None
        """
        args = self.args
        source("..")
        AtomType.clear_type()
        ResidueType.clear_type()
        for frc in GlobalSetting.BondedForces:
            frc.clear_type()
        source("..forcefield.charmm27.protein")
        source("__main__")

        t = ACE + ALA * 10 + NME
        Save_SPONGE_Input(t, f"{args.o}")

    def test_assign(self):
        """
        This **function** does the assignment test for Xponge

        :param args: arguments from argparse
        :return: None
        """
        args = self.args
        source("..")
        source("..forcefield.amber.gaff")
        source("__main__")

        t = assign.Assign()
        t.add_atom("O", 0, 0, 0)
        t.add_atom("H", 1, 0, 0)
        t.add_atom("H", 0, 1, 0)
        t.add_bond(0, 1, 1)
        t.add_bond(0, 2, 1)
        t.determine_ring_and_bond_type()
        t.determine_atom_type("gaff")
        equal_atoms = t.Determine_Equal_Atoms()
        t.calculate_charge("resp", opt=True, extra_equivalence=equal_atoms)
        Save_PDB(t, f"{args.o}.pdb")
        Save_Mol2(t, f"{args.o}.mol2")
        wat = t.to_residuetype("WAT")
        Save_PDB(wat, f"{args.o}_Residue.pdb")
        Save_Mol2(wat, f"{args.o}_Residue.mol2")
        Save_SPONGE_Input(wat, f"{args.o}")

        t = load_mol2(f"{args.o}_Residue.mol2")
        wat_ = t.residues[0]
        diff = abs(wat_.O.charge + 0.8)
        self.assertLess(diff, 0.02)
        diff = abs(wat_.H.charge - 0.4)
        self.assertLess(diff, 0.02)
        diff = abs(wat_.H.charge - wat_.H1.charge)
        self.assertLess(diff, 0.01)

    def test_lattice(self):
        """
        This **function** does the assignment test for the lattice functions

        :param args: arguments from argparse
        :return: None
        """
        args = self.args
        source("..")
        source("..forcefield.amber.tip3p")
        source("__main__")
        box = BlockRegion(0, 0, 0, 60, 60, 60)
        region_1 = BlockRegion(0, 0, 20, 20, 20, 40)
        region_2 = BlockRegion(0, 0, 40, 20, 20, 60)
        region_3 = BlockRegion(0, 0, 0, 20, 20, 20)
        region_4 = SphereRegion(20, 10, 30, 10)
        region_5 = BlockRegion(0, 0, 0, 20, 20, 40, side="out")
        region_2or3 = UnionRegion(region_2, region_3)
        region_4and5 = IntersectRegion(region_4, region_5)
        region_6 = FrustumRegion(10, 40, 0, 15, 10, 40, 60, 1)
        region_7 = PrismRegion(30, 30, 0, 20, 0, 0, 0, 20, 0, 10, 10, 20)
        t = Lattice("bcc", basis_molecule=CL, scale=4)
        t2 = Lattice("fcc", basis_molecule=K, scale=3)
        t3 = Lattice("sc", basis_molecule=NA, scale=3)
        t4 = Lattice("hcp", basis_molecule=MG2, scale=4)
        t5 = Lattice("diamond", basis_molecule=AL3, scale=5)
        mol = t.Create(box, region_1)
        mol = t2.create(box, region_2or3, mol)
        mol = t3.create(box, region_4and5, mol)
        mol = t4.create(box, region_6, mol)
        mol = t5.create(box, region_7, mol)
        Save_PDB(mol, f"{args.o}.pdb")

    def test_fep(self):
        """
        This **function** does the assignment test for the FEP functions

        :param args: arguments from argparse
        :return: None
        """
        source("..")
        source("..forcefield.amber.ff14sb")
        source("__main__")
        args = self.args
        t = NALA | ACE + ALA + NME
        for atom in t.residues[0].atoms:
            atom.x += 10
        Save_PDB(t, f"{args.o}.pdb")
        Save_Mol2(NALA, f"{args.o}_r1.mol2")
        Save_Mol2(NGLY, f"{args.o}_r2.mol2")
        error = os.system(f"Xponge mol2rfe -nl 1 -pdb {args.o}.pdb -r1 {args.o}_r1.mol2 \
-r2 {args.o}_r2.mol2 -pstep 1000 -esteps 1000 > {os.devnull}")
        self.assertEqual(error, 0)


def _one_test(ccon, name, args):
    """

    :param ccon:
    :param name:
    :param args:
    :return:
    """
    devnull = Xopen(os.devnull, 'w')
    sys.stdout = devnull
    sys.stderr = devnull
    runner = unittest.TextTestRunner(stream=devnull)
    t = runner.run(TestMyPackage.get_test_suite(name, args))
    ccon.send([t.errors, t.failures])
    devnull.close()
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


def test(args):
    """
    This **function** does the tests for Xponge

    :param args: arguments from argparse
    :return: None
    """
    GlobalSetting.verbose = args.verbose
    if not args.do:
        args.do = [["base"]]
    args.do = args.do[0]
    if "all" in args.do:
        args.do = ["base", "charmm27", "assign", "lattice"]
    fcon, ccon = mpc.Pipe()
    errors = []
    failures = []
    Xprint("Test(s): " + " ".join(args.do), verbose=-1)
    Xprint("======================================================", verbose=-1)
    for name in args.do:
        p = mpc.Process(target=_one_test, args=(ccon, name, args))
        p.start()
        p.join()
        new_errors, new_failures = fcon.recv()
        errors.extend(new_errors)
        failures.extend(new_failures)
        this_correct = f"{len(new_errors)} error(s) and {len(new_failures)} failure(s)"
        for ei, error in enumerate(new_errors):
            Xprint(f"Error {ei+1} for {name}", verbose=1)
            Xprint(f"{error[1]}", verbose=1)
        for ei, error in enumerate(new_failures):
            Xprint(f"Failure {ei+1} for {name}", verbose=1)
            Xprint(f"{error[1]}", verbose=1)
        Xprint(f"{name}: {this_correct}", verbose=-1)
    Xprint("======================================================", verbose=-1)
    if not errors and not failures:
        Xprint("No error or failure", verbose=-1)
    else:
        Xprint(f"{len(errors)} error(s) and {len(failures)} failure(s) found", verbose=-1)
        sys.exit(len(errors) + len(failures))


def converter(args):
    """
    This **function** converts the format of coordinate file

    :param args: arguments from argparse
    :return: None
    """
    from ..analysis import md_analysis as xmda
    import MDAnalysis as mda

    if args.c:
        if args.cf == "guess":
            u = mda.Universe(args.p, args.c)
        elif args.cf == "sponge_crd":
            u = mda.Universe(args.p, args.c, format=xmda.SpongeCoordinateReader)
        elif args.cf == "sponge_traj":
            dirname, basename = os.path.split(args.c)
            if basename == "mdcrd.dat":
                box = "mdbox.txt"
            else:
                box = basename.replace(".dat", ".box")
            box = os.path.join(dirname, box)
            if box and os.path.exists(box):
                u = mda.Universe(args.p, args.c, box=box, format=xmda.SpongeTrajectoryReader)
            else:
                u = mda.Universe(args.p, args.c, format=xmda.SpongeTrajectoryReader)
    else:
        u = mda.Universe(args.p)

    if args.of == "sponge_crd":
        with xmda.SpongeCoordinateWriter(args.o) as w:
            w.write(u)
    elif args.of == "sponge_traj":
        with xmda.SpongeTrajectoryWriter(args.o) as w:
            for _ in u.trajectory:
                w.write(u)
    else:
        with mda.Writer(args.o, n_atoms=len(u.coord.positions)) as w:
            for _ in u.trajectory:
                w.write(u)


def maskgen(args):
    """
    This **function** uses VMD to generate mask

    :param args: arguments from argparse
    :return: None
    """
    s = input("Please Enter Your Selection Mask:\n")

    p = args.p.split(os.path.sep)
    p = "/".join(p)

    c = ""
    if args.c:
        c = args.c.split(os.path.sep)
        c = "/".join(c)
        c = "mol addfile " + c

    temp_write = """set f [open "{0}" "w"]
mol new {1}
{2}
atomselect top "{3}"
puts $f [atomselect0 list]
close $f
quit
""".format(args.o, p, c, s)

    temp = Xopen("maskgen_temp_tcl_file", "w")
    temp.write(temp_write)
    temp.close()

    os.system("{0} -dispdev none -e maskgen_temp_tcl_file".format(args.vmd))
    os.remove("maskgen_temp_tcl_file")


def exgen(args):
    """
    This **function** reads the SPONGE input files for bonded interactions and generate a exclude file

    :param args: arguments from argparse
    :return: None
    """
    partners = [set([]) for i in range(args.n)]

    def exclude_2_atoms(words):
        i, j = int(words[0]), int(words[1])
        partners[i].add(j)
        partners[j].add(i)

    def exclude_3_atoms(words):
        i, k = int(words[0]), int(words[2])
        partners[i].add(k)
        partners[k].add(i)

    def exclude_4_atoms(words):
        i, l = int(words[0]), int(words[3])
        partners[i].add(l)
        partners[l].add(i)

    for bond in args.bond:
        with open(bond) as f:
            f.readline()
            for line in f:
                words = line.split()
                exclude_2_atoms(words)

    for angle in args.angle:
        with open(angle) as f:
            f.readline()
            for line in f:
                words = line.split()
                exclude_3_atoms(words)
    for dihedral in args.dihedral:
        with open(dihedral) as f:
            f.readline()
            for line in f:
                words = line.split()
                exclude_4_atoms(words)

    for virtual in args.virtual:
        with open(virtual) as f:
            for line in f:
                words = line.split()
                t = int(words[0])
                if t == 0:
                    exclude_2_atoms(words[1:])
                elif t == 1:
                    exclude_3_atoms(words[1:])
                elif t in (2, 3):
                    exclude_4_atoms(words[1:])
                else:
                    raise Exception("virtual atom type wrong: are you sure this is a SPONGE virtual atom file?")

    for exclude in args.exclude:
        with open(exclude) as f:
            f.readline()
            count = 0
            for line in f:
                words = line.split()
                t = set(words[1:])
                partners[count] = partners[count].union(t)
                count += 1

    total = 0
    towrite = "{} {}\n"
    for i, p in enumerate(partners):
        newp = []
        for pi in p:
            if pi > i:
                newp.append(pi)
        towrite += "%d " % len(newp)
        towrite += ("{} " * len(newp)).format(*newp) + "\n"
        total += len(newp)
        towrite = towrite.format(args.n, total)

    f = Xopen(args.o, "w")
    f.write(towrite)
    f.close()


def name2name(args):
    """
    This **function** change the atom names from one file to another file

    :param args: arguments from argparse
    :return: None
    """
    from rdkit import Chem
    from rdkit.Chem import rdFMCS
    source("..")
    rdktool = source("..helper.rdkit")

    if args.to_format == "mol2":
        to_ = assign.Get_Assignment_From_Mol2(args.to_file)
    elif args.to_format == "gaff_mol2":
        source("..forcefield.amber.gaff")
        to_ = load_mol2(args.to_file).residues[0]
        to_ = assign.Get_Assignment_From_ResidueType(to_)
    elif args.to_format == "pdb":
        to_ = assign.Get_Assignment_From_PDB(args.to_file, determine_bond_order=False,
                                             only_residue=args.to_residue)

    if args.from_format == "mol2":
        from_ = assign.Get_Assignment_From_Mol2(args.from_file)
    elif args.from_format == "gaff_mol2":
        source("..forcefield.amber.gaff")
        from_ = load_mol2(args.from_file).residues[0]
        from_ = assign.Get_Assignment_From_ResidueType(from_)
    elif args.from_format == "pdb":
        from_ = assign.Get_Assignment_From_PDB(args.from_file, determine_bond_order=False,
                                               only_residue=args.from_residue)

    rdmol_a = rdktool.assign_to_rdmol(to_, True)
    rdmol_b = rdktool.assign_to_rdmol(from_, True)

    result = rdFMCS.FindMCS([rdmol_a, rdmol_b], completeRingsOnly=True, timeout=args.tmcs)
    rdmol_mcs = Chem.MolFromSmarts(result.smartsString)

    match_a = rdmol_a.GetSubstructMatch(rdmol_mcs)
    match_b = rdmol_b.GetSubstructMatch(rdmol_mcs)
    matchmap = {from_.names[match_b[j]]: to_.names[match_a[j]] for j in range(len(match_a))}
    from_.names = [matchmap.get(name, name) for name in from_.names]
    from_.name = args.out_residue

    if args.out_format == "mol2":
        from_.Save_As_Mol2(args.out_file)
    elif args.out_format == "pdb":
        from_.Save_As_PDB(args.out_file)
    elif args.out_format == "mcs_pdb":
        towrite = towrite = "REMARK   Generated By Xponge (Max Common Structure)\n"
        for i, atom in enumerate(from_.atoms):
            if i in match_b:
                towrite += "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f%17s%2s\n" % (i + 1, from_.names[i],
                                                                                     from_.name, " ", 1,
                                                                                     from_.coordinate[i][0],
                                                                                     from_.coordinate[i][1],
                                                                                     from_.coordinate[i][2], " ", atom)
        f = Xopen(args.out_file, "w")
        f.write(towrite)
        f.close()


def _mol2rfe_build(args, merged_from, merged_to):
    """

    :param args:
    :param merged_from:
    :param merged_to:
    :return:
    """
    source("..")
    fep = source("..forcefield.special.fep", False)

    if "build" in args.do:
        Xprint("\nBUILDING TOPOLOGY\n", verbose=-1)
        fep.Save_Soft_Core_LJ()

        for i in range(args.nl + 1):
            if os.path.exists("%d" % i):
                os.system("rm -rf %d" % i)
            os.mkdir("%d" % i)
            tt = fep.Merge_Force_Field(merged_from, merged_to, i / args.nl)
            if i == 0:
                Xprint("Initial Structure Optimizing\n")
                process.optimize(tt,
                                 only_bad_coordinate=False,
                                 extra_commands={"lambda_lj": i / args.nl})
                for atom_m, atom_t in zip(merged_from.atoms, tt.atoms):
                    atom_m.x = atom_t.x
                    atom_m.y = atom_t.y
                    atom_m.z = atom_t.z
            build.save_mol2(tt, "%d/%s.mol2" % (i, args.temp))
            build.Save_SPONGE_Input(tt, "%d/%s" % (i, args.temp))
            Xprint(f"{i} built success")

def _mol2rfe_output_path(subdir, workdir, tempname):
    """

    :param subdir:
    :param workdir:
    :param tempname:
    :return:
    """
    toadd = " -mdinfo {2}/{0}/{1}.mdinfo -mdout {2}/{0}/{1}.mdout".format(subdir, tempname, workdir)
    toadd += " -rst {2}/{0}/{1} -crd {2}/{0}/{1}.dat -box {2}/{0}/{1}.box".format(subdir, tempname, workdir)
    return toadd


def _mol2rfe_min(args):
    """

    :param args:
    :return:
    """
    source("..")

    if "min" in args.do:
        for i in range(args.nl + 1):
            if os.path.exists("%d/min" % i):
                os.system("rm -rf %d/min" % i)
            os.mkdir("%d/min" % i)
            basic = f"SPONGE -default_in_file_prefix {i}/{args.temp}"
            lambda_ = i / args.nl
            basic += f" -mode minimization -lambda_lj {lambda_}"
            basic += _mol2rfe_output_path("min", i, args.temp)
            if not args.mi:
                cif = " -constrain_mode SHAKE -cutoff 8"
                run(f"{basic} {cif} -dt 1e-8 -step_limit {args.msteps[0]}")
                cif += " -coordinate_in_file {1}/min/{0}_coordinate.txt".format(args.temp, i)
                run(f"{basic} {cif} -dt 1e-7 -step_limit {args.msteps[1]}")
                run(f"{basic} {cif} -dt 1e-6 -step_limit {args.msteps[2]}")
                run(f"{basic} {cif} -dt 1e-5 -step_limit {args.msteps[3]}")
                run(f"{basic} {cif} -dt 1e-4 -step_limit {args.msteps[4]}")
                run(f"{basic} {cif} -dt 1e-3 -step_limit {args.msteps[5]}")
            else:
                mdin = args.mi.pop(0)
                cif = ""
                run(f"{basic} {cif} -mdin {mdin}")
                cif += " -coordinate_in_file {1}/min/{0}_coordinate.txt".format(args.temp, i)
                for mdin in args.mi:
                    run(f"{basic} {cif} -mdin {mdin}")


def _mol2rfe_pre_equilibrium(args):
    """

    :param args:
    :return:
    """
    source("..")

    if "pre_equilibrium" in args.do:
        for i in range(args.nl + 1):
            if os.path.exists("%d/pre_equilibrium" % i):
                os.system("rm -rf %d/pre_equilibrium" % i)
            os.mkdir("%d/pre_equilibrium" % i)
            command = f"SPONGE -default_in_file_prefix {i}/{args.temp}"
            lambda_ = i / args.nl
            command += f" -lambda_lj {lambda_} -cutoff 8"
            command += _mol2rfe_output_path("pre_equilibrium", i, args.temp)
            command += f" -coordinate_in_file {i}/min/{args.temp}_coordinate.txt"
            if not args.pi:
                command += f" -mode NPT -step_limit {args.pre_equilibrium_step} -dt {args.dt} -constrain_mode SHAKE"
                command += f" -barostat {args.barostat} -thermostat {args.thermostat}"
                run(command)
            else:
                command += f" -mdin {args.pi}"
                run(command)


def _mol2rfe_equilibrium(args):
    """

    :param args:
    :return:
    """
    source("..")

    if "equilibrium" in args.do:
        for i in range(args.nl + 1):
            if os.path.exists("%d/equilibrium" % i):
                os.system("rm -rf %d/equilibrium" % i)
            os.mkdir("%d/equilibrium" % i)
            command = f"SPONGE -default_in_file_prefix {i}/{args.temp}"
            lambda_ = i / args.nl
            command += f" -lambda_lj {lambda_} -cutoff 8"
            command += _mol2rfe_output_path("equilibrium", i, args.temp)
            command += f" -coordinate_in_file {i}/pre_equilibrium/{args.temp}_coordinate.txt"
            if not args.ei:
                command += f" -mode NPT -step_limit {args.equilibrium_step} -dt {args.dt} -constrain_mode SHAKE"
                command += f" -barostat {args.barostat} -thermostat {args.thermostat}"
                command += f" -write_information_interval 100 -write_restart_file_interval {args.equilibrium_step}"
                run(command)
            else:
                command += f" -mdin {args.pi}"
                run(command)


def _mol2rfe_analysis(args, merged_from, merged_to):
    """

    :param args:
    :param merged_from:
    :param merged_to:
    :return:
    """
    source("..")
    source("..analysis")

    if "analysis" in args.do:
        f = Xopen("dh_dlambda.txt", "w")
        f.close()
        resname = merged_from.residues[args.ri].name
        draw_r1_mol = merged_from.deepcopy()
        draw_r2_mol = merged_to.deepcopy()
        load_coordinate(f"0/equilibrium/{args.temp}_coordinate.txt", draw_r1_mol)
        load_coordinate(f"{args.nl}/equilibrium/{args.temp}_coordinate.txt", draw_r2_mol)
        draw_r1_res = draw_r1_mol.residues[args.ri]
        draw_r2_res = draw_r2_mol.residues[args.ri]
        draw_r1_res.name = resname.split("_")[0]
        draw_r2_res.name = resname.split("_")[1]
        to_delete = []
        for atom in draw_r1_res.atoms:
            if atom.LJtype == "ZERO_LJ_ATOM":
                to_delete.append(atom)
        for atom in to_delete:
            draw_r1_res.atoms.remove(atom)
        to_delete = []
        for atom in draw_r2_res.atoms:
            if atom.LJtype == "ZERO_LJ_ATOM":
                to_delete.append(atom)
            if atom.name.endswith("R2"):
                atom.name = atom.name[:-2]
        for atom in to_delete:
            draw_r2_res.atoms.remove(atom)
        save_pdb(draw_r1_mol, "r1.pdb")
        save_pdb(draw_r2_mol, "r2.pdb")
        if args.method == "TI":
            ti = source(".ti", False)
            ti.ti_analysis(args, merged_from)
        elif args.method == "FEP_BAR":
            raise NotImplementedError


def mol2rfe(args):
    """
    This **function** helps with the relative free energy calculation

    :param args: arguments from argparse
    :return: None
    """
    source("..")
    source("..forcefield.special.fep")
    source("..forcefield.special.min")

    if not args.ff:
        source("..forcefield.amber.gaff")
        source("..forcefield.amber.ff14sb")
        source("..forcefield.amber.tip3p")
    else:
        idic, ipy = os.path.split(args.ff)
        sys.path.append(idic)
        ipy, isuffix = os.path.splitext(ipy)
        assert isuffix == ".py", "the input force field file should be an Xponge file written by python"
        __import__(ipy)

    if not args.do:
        args.do = [["build", "min", "pre_equilibrium", "equilibrium", "analysis"]]
    args.do = args.do[0]

    from_res_type_ = load_mol2(args.r1).residues[0]
    from_ = assign.Get_Assignment_From_ResidueType(from_res_type_)
    if not args.ff:
        parmchk2_gaff(args.r1, args.temp + "_TMP1.frcmod")

    to_res_type_ = load_mol2(args.r2).residues[0]
    to_ = assign.Get_Assignment_From_ResidueType(to_res_type_)
    if not args.ff:
        parmchk2_gaff(args.r2, args.temp + "_TMP2.frcmod")

    for mol2file in args.r0:
        load_mol2(mol2file)

    rmol = load_pdb(args.pdb)

    merged_from, merged_to = Merge_Dual_Topology(rmol, rmol.residues[args.ri],
                                                 to_res_type_, from_, to_,
                                                 args.tmcs, f"{args.fmcs}", args.lmcs)

    if args.dohmr:
        H_Mass_Repartition(merged_from)
        H_Mass_Repartition(merged_to)

    _mol2rfe_build(args, merged_from, merged_to)

    _mol2rfe_min(args)

    _mol2rfe_pre_equilibrium(args)

    _mol2rfe_equilibrium(args)

    _mol2rfe_analysis(args, merged_from, merged_to)



"""
This **module** is used to calculate the Gasteiger charge.
"""
from rdkit.Chem import rdPartialCharges
from ..helper import Xprint, set_global_alternative_names
from ..helper.rdkit import assign_to_rdmol


def gasteiger(assign):
    """
    This **function** is used to calculate the Gasteiger charge of an assignment

    :param assign: the Assign instance
    :return: a list of charges
    """
    rdmol = assign_to_rdmol(assign)
    rdPartialCharges.ComputeGasteigerCharges(rdmol)
    return [float(atom.GetProp("_GasteigerCharge")) for atom in rdmol.GetAtoms()]


set_global_alternative_names()

Xprint("""Reference for Gasteiger charge
  Gasteiger, J., Marsili, M.
    A new model for calculating atomic charges in molecules
    Tetrahedron Letters, 1978 19(34) 3181-3184
    DOI: 10.1016/S0040-4039(01)94977-9""")



"""
This **module** is the basic setting for the force field format of atom-specific cmap
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, Xdict, set_global_alternative_names

# pylint: disable=invalid-name
CMapType = Generate_New_Bonded_Force_Type("cmap", "1-2-3-4-5", {"resolution": int, "parameters": list},
                                          False)

@CMapType.Set_Same_Force_Function
def cmap_same_force(_, atom_list):
    """
    This **function** is used to return the same force type for an atom list
    :param _:
    :param atom_list:
    :return:
    """
    return [atom_list]


@Molecule.Set_Save_SPONGE_Input("cmap")
def write_cmap(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    bonds = []
    haved_cmaps = []
    haved_cmap_index = Xdict()
    for bond in self.bonded_forces.get("cmap", []):
        if bond.type not in haved_cmaps:
            haved_cmap_index[bond.type] = len(haved_cmaps)
            haved_cmaps.append(bond.type)
        bonds.append("%d %d %d %d %d %d" % (self.atom_index[bond.atoms[0]]
                                            , self.atom_index[bond.atoms[1]],
                                            self.atom_index[bond.atoms[2]],
                                            self.atom_index[bond.atoms[3]],
                                            self.atom_index[bond.atoms[4]], haved_cmap_index[bond.type]))

    if bonds:
        towrite = "%d %d\n" % (len(bonds), len(haved_cmaps))
        for bondtype in haved_cmaps:
            towrite += "%d " % bondtype.resolution
        towrite += "\n"
        for bondtype in haved_cmaps:
            for i, pi in enumerate(bondtype.parameters):
                towrite += "%f " % pi
                if (i + 1) % bondtype.resolution == 0:
                    towrite += "\n"
            towrite += "\n"
        bonds.sort(key=lambda x: list(map(int, x.split()[:5])))
        towrite += "\n".join(bonds)

        return towrite
    return None

set_global_alternative_names()



"""
This **module** set the basic configuration for ff14sb
"""
from ...helper import source, Xprint, set_real_global_variable

source("....")
amber = source("...amber")

AtomType.New_From_String(
    """
    name mass    charge[e]      LJtype
    HW   1.008   0.52000        HW
    OW   16      0              OW
    EP   0      -1.04000        EPW
    """)

bond_base.BondType.New_From_String(r"""
name    k[kcal/mol·A^-2]   b[A]
OW-HW   553                0.9572
HW-HW   553                1.5136
""")

angle_base.AngleType.New_From_String(r"""
name        k       b
HW-OW-HW    0       0
OW-HW-HW    0       0
""")

lj_base.LJType.New_From_String(r"""
name    epsilon[kcal/mol]   sigma[A]
OW-OW   0.155               3.154
HW-HW   0                   0
EPW-EPW 0                   0
""")

virtual_atom_base.VirtualType2.New_From_String(r"""
name   atom0    atom1   atom2   k1         k2
EP     -3       -2      -1      0.1279696  0.1279696
""")

TIP4P = load_mol2(os.path.join(AMBER_DATA_DIR, "tip4pew.mol2"), as_template=True)

ResidueType.set_type("H2O", ResidueType.get_type("WAT"))
ResidueType.set_type("HOH", ResidueType.get_type("WAT"))

set_real_global_variable("H2O", ResidueType.get_type("WAT"))
set_real_global_variable("HOH", ResidueType.get_type("WAT"))

Xprint("""Reference for tip4p:
  William L. Jorgensen, Jayaraman Chandrasekhar, Jeffry D. Madura, Roger W. Impey and Michael L. Klein
    Comparison of simple potential functions for simulating liquid water
    The Journal of Chemical Physics 1983, 79, 926-935
    DOI: 10.1063/1.445869

""")
# pylint:disable=undefined-variable



"""
This **module** helps to assign bond orders
"""
from itertools import product
from . import AssignRule, Xdict, OrderedDict, set_attribute_alternative_names, deepcopy, np

bo = AssignRule("bo", pure_string=True)

@bo.add_rule("X")
def _(i, assign):
    return assign.atoms[i] in ("H", "F", "Cl", "Br", "I")


@bo.add_rule("Ccn")
def _(i, assign):
    return assign.Atom_Judge(i, "C1") and [assign.atoms[key] for key in assign.bonds[i].keys()][0] == "N"


@bo.add_rule("Cx1")
def _(i, assign):
    return assign.Atom_Judge(i, "C1")


@bo.add_rule("Co2")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "C3") and bonded_os == 2


@bo.add_rule("C")
def _(i, assign):
    return assign.atoms[i] == "C"


@bo.add_rule("Nnn1")
def _(i, assign):
    return assign.Atom_Judge(i, "N1") and [assign.atoms[key] for key in assign.bonds[i].keys()][0] == "N"


@bo.add_rule("Nx1")
def _(i, assign):
    return assign.Atom_Judge(i, "N1")


@bo.add_rule("Nnn2")
def _(i, assign):
    return assign.Atom_Judge(i, "N2") and "N" in {assign.atoms[key] for key in assign.bonds[i].keys()}


@bo.add_rule("Nx2")
def _(i, assign):
    return assign.Atom_Judge(i, "N2")


@bo.add_rule("No2")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "N3") and bonded_os == 2


@bo.add_rule("No1")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "N3") and bonded_os == 1


@bo.add_rule("Nx3")
def _(i, assign):
    return assign.Atom_Judge(i, "N3")


@bo.add_rule("Nx4")
def _(i, assign):
    return assign.Atom_Judge(i, "N4")


@bo.add_rule("On1")
def _(i, assign):
    return assign.Atom_Judge(i, "O1") and [assign.atoms[key] for key in assign.bonds[i].keys()].count("N") == 1


@bo.add_rule("Ox1")
def _(i, assign):
    return assign.Atom_Judge(i, "O1")


@bo.add_rule("Ox2")
def _(i, assign):
    return assign.Atom_Judge(i, "O2")


@bo.add_rule("Px1")
def _(i, assign):
    return assign.Atom_Judge(i, "P1")


@bo.add_rule("Px2")
def _(i, assign):
    return assign.Atom_Judge(i, "P2")


@bo.add_rule("Px3")
def _(i, assign):
    return assign.Atom_Judge(i, "P3")


@bo.add_rule("Po3")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "P4") and bonded_os == 3



@bo.add_rule("Po2")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "P4") and bonded_os == 2


@bo.add_rule("Px4")
def _(i, assign):
    return assign.Atom_Judge(i, "P4")


@bo.add_rule("Sn1")
def _(i, assign):
    return assign.Atom_Judge(i, "S1") and [assign.atoms[key] for key in assign.bonds[i].keys()].count("N") == 1


@bo.add_rule("Sx1")
def _(i, assign):
    return assign.Atom_Judge(i, "S1")


@bo.add_rule("Sx2")
def _(i, assign):
    return assign.Atom_Judge(i, "S2")


@bo.add_rule("S3")
def _(i, assign):
    return assign.Atom_Judge(i, "S3")


@bo.add_rule("So3")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "S4") and bonded_os >= 3


@bo.add_rule("So2")
def _(i, assign):
    bonded_os = sum(assign.Atom_Judge(key, "O1") or assign.Atom_Judge(key, "S1") for key in assign.bonds[i].keys())
    return assign.Atom_Judge(i, "S4") and bonded_os == 2


@bo.add_rule("Sx4")
def _(i, assign):
    return assign.Atom_Judge(i, "S4")


class BondOrderAssignment:
    """
    This **class** includes the functions to assign bond orders

    :param original_penalties: the original penalties dict
    :param max_stat: the max valence stats to iterate
    :param assign: the father Assignment instance
    """
    atomic_valence = Xdict({
        "X": OrderedDict({0: 64, 1: 0, 2: 64}),
        "Cn1": OrderedDict({3: 0, 4: 1, 5: 32}),
        "Cx1": OrderedDict({3: 1, 4: 0, 5: 32}),
        "Co2": OrderedDict({4: 32, 5: 0, 6: 32}),
        "C": OrderedDict({2: 64, 3: 32, 4: 0, 5: 32, 6: 64}),
        "Nnn1": OrderedDict({2: 0, 3: 0}),
        "Nx1": OrderedDict({2: 3, 3: 0, 4: 32}),
        "Nnn2": OrderedDict({3: 1, 4: 0}),
        "Nx2": OrderedDict({2: 4, 3: 0, 4: 32}),
        "No2": OrderedDict({3: 64, 4: 32, 5: 0, 6: 32}),
        "No1": OrderedDict({3: 1, 4: 0}),
        "Nx3": OrderedDict({2: 32, 3: 0, 4: 1, 5: 2}),
        "Nx4": OrderedDict({3: 64, 4: 0, 5: 64}),
        "On1": OrderedDict({1: 0, 0: 1}),
        "Ox1": OrderedDict({2: 0, 1: 1, 3: 64}),
        "Ox2": OrderedDict({1: 32, 2: 0, 3:64}),
        "Px1": OrderedDict({2: 2, 3: 0, 4: 32}),
        "Px2": OrderedDict({2: 4, 3: 0, 4: 2}),
        "Px3": OrderedDict({3: 32, 4: 0, 5: 1, 6: 2}),
        "Po2": OrderedDict({5: 32, 6: 0, 7: 32}),
        "Po3": OrderedDict({6: 32, 7: 0}),
        "Px4": OrderedDict({3: 64, 4: 1, 5: 0, 6: 32}),
        "Sn1": OrderedDict({1: 0, 2: 1}),
        "Sx1": OrderedDict({1: 2, 2: 0, 3: 64}),
        "Sx2": OrderedDict({1: 32, 2: 0, 3: 32}),
        "Sx3": OrderedDict({3: 1, 4: 0, 5: 2, 6: 2}),
        "So2": OrderedDict({6: 0, 7: 32}),
        "So3": OrderedDict({6: 32, 7: 0}),
        "Sx4": OrderedDict({4: 4, 5: 2, 6:0}),
    })
    def __init__(self, original_penalties, max_step, max_stat, assign, debug=False):
        self.debug = debug
        self.max_step = max_step
        self.max_stat = max_stat
        self.assign = assign
        self.original_penalties = original_penalties
        self.uc = [set(filter(lambda aj: self.assign.bonds[ai][aj] == -1, self.assign.bonds[ai]))
                     for ai in range(self.assign.atom_numbers)]
        self.connected = [sum(filter(lambda x: x > 0, self.assign.bonds[ai].values()))
                     for ai in range(self.assign.atom_numbers)]
        self.valence_best = [next(iter(pi)) - self.connected[i] for i, pi in enumerate(self.original_penalties)]
        self.valence = deepcopy(self.valence_best)
        self.penalties = Xdict(not_found_method=lambda x: [])
        self._get_penalties(original_penalties)
        self.stat_position = 0
        self.current_stat = 0
        self.points = []
        self.cached = {}
        set_attribute_alternative_names(self)

    def _get_penalties(self, original_penalties):
        """

        :param original_penalties:
        :return:
        """
        for atom, pi in enumerate(original_penalties):
            for valence, penalty in pi.items():
                self.penalties[penalty].append((atom, penalty, valence))

    def _preprocess_penalties(self, n):
        if n in self.cached:
            return self.cached[n]
        if n == 1:
            toret = [[point] for point in self.penalties[n]]
        else:
            toret =  []
            have_added = set()
            for i in range(1, n // 2 + 1):
                r1 = self.cached[i]
                r2 = self.cached[n - i]
                for ri, rj in product(r1, r2):
                    if {r0[0] for r0 in ri} & {r0[0] for r0 in rj}:
                        continue
                    rij = ri + rj
                    rij.sort()
                    rij_checkstring = "+".join([f"{r0[0]}-{r0[1]}-{r0[2]}" for r0 in rij])
                    if rij_checkstring not in have_added:
                        have_added.add(rij_checkstring)
                        toret.append(rij)
            toret.extend([[point] for point in self.penalties.get(n, [])])
        self.cached[n] = toret
        return toret

    def _get_next_valence(self):
        if self.current_stat < self.max_stat:
            if self.stat_position < len(self.points):
                self.valence = deepcopy(self.valence_best)
                has_negative_value = False
                for point in self.points[self.stat_position]:
                    self.valence[point[0]] = point[2] - self.connected[point[0]]
                    if self.valence[point[0]] < 0:
                        has_negative_value = True
                        break
                self.stat_position += 1
                if has_negative_value:
                    return True
            else:
                self.current_stat += 1
                self.stat_position = 0
                self.points = self._preprocess_penalties(self.current_stat)
                if self.debug:
                    print("stat=", self.current_stat)
                    print("points=\n", self.points)
                return True


    def main(self):
        """
        This **function** is the main function to do the bond order assignment

        :return: True for success, False for failure
        """
        count = 0
        success = False
        while count < self.max_step and self.current_stat < self.max_stat and not success:
            bonds = deepcopy(self.assign.bonds)
            uc = deepcopy(self.uc)
            valence = deepcopy(self.valence)
            guess_bonds = []
            determined = False
            while not determined:
                index_sort = np.argsort([len(uci) for uci in uc]).tolist()
                no_basic_rule = True
                for i in index_sort:
                    if self.debug:
                        print(i, valence, uc)
                    if len(uc[i]) == valence[i] and valence[i] != 0:
                        while uc[i]:
                            j = uc[i].pop()
                            bonds[i][j] = 1
                            bonds[j][i] = 1
                            uc[j].remove(i)
                            valence[j] -= 1
                        valence[i] = 0
                        no_basic_rule = False
                    elif len(uc[i]) == 1:
                        j = uc[i].pop()
                        uc[j].remove(i)
                        bonds[j][i] = valence[i]
                        bonds[i][j] = valence[i]
                        valence[j] -= valence[i]
                        valence[i] -= valence[i]
                        no_basic_rule = False
                success = True
                for i in range(self.assign.atom_numbers):
                    if valence[i] != 0 or len(uc[i]) != 0:
                        success = False
                    if len(uc[i]) == 0 and valence[i] != 0 or valence[i] == 0 and len(uc[i]) != 0:
                        success = False
                        determined = True
                        break
                if success:
                    determined = True
                if not determined and no_basic_rule:
                    if not guess_bonds:
                        i = index_sort.pop()
                        j = uc[i].pop()
                        uc[j].remove(i)
                        guess_bonds = [i, j]
                        bonds[i][j] = 1
                        bonds[j][i] = 1
                        valence[i] -= 1
                        valence[j] -= 1
                    else:
                        i, j = guess_bonds
                        if bonds[i][j] == 3:
                            determined = True
                            success = False
                            break
                        else:
                            bonds[i][j] += 1
                            bonds[j][i] += 1
                            valence[i] -= 1
                            valence[j] -= 1
            if self.debug:
                print(valence, uc)
            if not success:
                count += 1
                while self._get_next_valence():
                    pass
                if self.debug:
                    print("-"*20, self.points[self.stat_position - 1])
        if success:
            self.assign.bonds = bonds
        return success

for key, value in BondOrderAssignment.atomic_valence.items():
    BondOrderAssignment.atomic_valence[key] = OrderedDict(sorted(value.items(), key=lambda t: t[1]))



"""
This **module** is the basic setting for the force field format of harmonic angle
"""
from ... import Generate_New_Bonded_Force_Type
from ...helper import Molecule, set_global_alternative_names, Xdict

# pylint: disable=invalid-name
AngleType = Generate_New_Bonded_Force_Type("angle", "1-2-3", {"k": float, "b": float}, True)

AngleType.Set_Property_Unit("k", "energy·angle^-2", "kcal/mol·rad^-2")
AngleType.Set_Property_Unit("b", "angle", "rad")


@Molecule.Set_Save_SPONGE_Input("angle")
def write_angle(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    angles = []
    for angle in self.bonded_forces.get("angle", []):
        order = list(range(3))
        if angle.k != 0:
            if self.atom_index[angle.atoms[order[0]]] > self.atom_index[angle.atoms[order[-1]]]:
                temp_order = order[::-1]
            else:
                temp_order = order
            angles.append("%d %d %d %f %f" % (self.atom_index[angle.atoms[temp_order[0]]]
                                              , self.atom_index[angle.atoms[temp_order[1]]],
                                              self.atom_index[angle.atoms[temp_order[2]]], angle.k, angle.b))

    if angles:
        towrite = "%d\n" % len(angles)
        angles.sort(key=lambda x: list(map(int, x.split()[:3])))
        towrite += "\n".join(angles)
        return towrite

    return None


#pylint: disable=unused-argument
@Molecule.Set_MindSponge_Todo("angle")
def _do(self, sys_kwarg, ene_kwarg, use_pbc):
    """

    :return:
    """
    from mindsponge.potential import AngleEnergy
    if "angle" not in ene_kwarg:
        ene_kwarg["angle"] = Xdict()
        ene_kwarg["angle"]["function"] = lambda system, ene_kwarg: AngleEnergy(
            index=ene_kwarg["angle"]["index"], use_pbc=use_pbc,
            force_constant=ene_kwarg["angle"]["force_constant"],
            bond_angle=ene_kwarg["angle"]["bond_angle"],
            energy_unit="kcal/mol")
        ene_kwarg["angle"]["index"] = []
        ene_kwarg["angle"]["force_constant"] = []
        ene_kwarg["angle"]["bond_angle"] = []
    bonds = []
    force_constants = []
    bond_angles = []
    for bond in self.bonded_forces.get("angle", []):
        if bond.k == 0:
            continue
        bonds.append([self.atom_index[atom] for atom in bond.atoms])
        force_constants.append(bond.k * 2)
        bond_angles.append(bond.b)
    ene_kwarg["angle"]["index"].append(bonds)
    ene_kwarg["angle"]["force_constant"].append(force_constants)
    ene_kwarg["angle"]["bond_angle"].append(bond_angles)

set_global_alternative_names()



"""
This **module** contains the functions for ti analysis
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from ..mdrun import run
from ..helper import Xopen
from ..analysis import MdoutReader


def ti_analysis(args, merged_from):
    """
    This **function** is used to do the ti analysis

    :param args: the arguments from the command line
    :return: None
    """
    ses = []
    prefix_sum = []
    suffix_sum = []
    frame = args.equilibrium_step // 100
    for i in range(args.nl + 1):
        if os.path.exists("%d/ti" % i):
            os.system("rm -rf %d/ti" % i)
        os.mkdir("%d/ti" % i)
        inprefix = f"{i}/{args.temp}"
        command = f"SPONGE_TI -LJ_soft_core_in_file {inprefix}_LJ_soft_core.txt"
        command += " -exclude_in_file {0}_exclude.txt -charge_in_file {0}_charge.txt".format(inprefix)
        command += f" -chargeA_in_file 0/{args.temp}_charge.txt"
        command += f" -chargeB_in_file {args.nl}/{args.temp}_charge.txt"
        lambda_ = i / args.nl
        command += f" -lambda_lj {lambda_}"
        command += f" -subsys_division_in_file {inprefix}_subsys_division.txt  -charge_pertubated 1"
        inprefix = f"{i}/ti/{args.temp}"
        command += f" -mdinfo {inprefix}.mdinfo -mdout {inprefix}.mdout"
        inprefix = f"{i}/equilibrium/{args.temp}"
        command += f" -crd {inprefix}.dat -box {inprefix}.box -TI dh_dlambda.txt"
        command += f" -atom_numbers {len(merged_from.atoms)}"
        command += f" -frame_numbers {frame}"
        if not args.ai:
            run(command)
        else:
            command += f" -mdin {args.ai}"
            run(command)
        temp = MdoutReader(f"{i}/ti/{args.temp}.mdout").dH_dlambda
        prefix_sum.append(np.cumsum(temp[::]) / (np.arange(frame) + 1))
        suffix_sum.append(np.cumsum(temp[::-1]) / (np.arange(frame) + 1))
        ses.append(np.std(temp) / np.sqrt(frame))
    prefix_sum = np.array(prefix_sum)
    dh_dlambda = np.loadtxt("dh_dlambda.txt")
    ses = np.array(ses)
    ses *= ses
    dh = []
    dh_int = []
    dh_se = []
    dh_int_se = []
    tempall = 0
    temp_se_all = 0
    space = 0.5 / args.nl
    if os.path.exists("time_check"):
        os.system("rm -rf time_check")
    os.mkdir("time_check")
    for i in range(args.nl):
        temp = (prefix_sum[i] + prefix_sum[i + 1]) * space
        time = args.dt * 0.1 * (np.arange(frame) + 1)
        plt.plot(time, temp, label="forward")
        temp_ses = np.vstack((time, temp))
        temp = (suffix_sum[i] + suffix_sum[i + 1]) * space
        temp_ses = np.vstack((temp_ses, temp))
        plt.plot(time, temp, label="backward")
        plt.xlabel("time[ns]")
        plt.ylabel("free energy difference[kcal/mol]")
        plt.legend()
        plt.savefig(f"time_check/{i}-{i + 1}.png")
        np.savetxt(f"time_check/{i}-{i + 1}.csv", temp_ses.transpose(),
                   header="time[ps],forward DeltaG[kcal/mol],backward DeltaG[kcal/mol]", comments="", delimiter=",")
        plt.clf()
        temp = dh_dlambda[i] * space
        temp += dh_dlambda[i + 1] * space
        temp_ses = space * (ses[i] + ses[i + 1])
        dh.append(temp)
        dh_se.append(np.sqrt(temp_ses))
        tempall += temp
        temp_se_all += temp_ses
        dh_int.append(tempall)
        dh_int_se.append(np.sqrt(temp_se_all))

    temp_ses **= 0.5
    f = Xopen("free_energy.txt", "w")
    f.write("lambda_state\tFE(i+1)-FE(i)[kcal/mol]\tFE(i+1)-FE(0)[kcal/mol]\n")
    f.write("\n".join(
        [f"{i}\t\t{dh[i]: .2f} +- {dh_se[i]:.2f}\t\t{dh_int[i]: .2f} +- {dh_int_se[i]:.2f}" for i in range(args.nl)]))
    f.close()



"""
This **module** is the basic setting for the force field property of charge
"""
import numpy as np
from ... import Generate_New_Pairwise_Force_Type
from ...helper import Molecule, AtomType, GlobalSetting, set_global_alternative_names, Xdict

AtomType.Add_Property({"LJtype": str})

# pylint: disable=invalid-name
LJType = Generate_New_Pairwise_Force_Type("LJ",
                                          {"epsilon": float, "rmin": float, "sigma": float, "A": float, "B": float})

LJType.Set_Property_Unit("rmin", "distance", "A")
LJType.Set_Property_Unit("sigma", "distance", "A")
LJType.Set_Property_Unit("epsilon", "energy", "kcal/mol")
LJType.Set_Property_Unit("A", "energy·distance^6", "kcal/mol·A^6")
LJType.Set_Property_Unit("B", "energy·distance^12", "kcal/mol·A^12")


@GlobalSetting.Add_Unit_Transfer_Function(LJType)
def _lj_unit_transfer(self):
    """
    This **function** is used to transfer the units of lj

    :param self:
    :return:
    """
    if self.A is not None and self.B is not None:
        if self.B == 0 or self.A == 0:
            self.sigma = 0
            self.epsilon = 0
        else:
            self.sigma = (self.A / self.B) ** (1 / 6)
            self.epsilon = 0.25 * self.B * self.sigma ** (-6)
        self.A = None
        self.B = None
    if self.sigma is not None:
        self.rmin = self.sigma * (4 ** (1 / 12) / 2)
        self.sigma = None


def lorentz_berthelot_for_a(epsilon1, rmin1, epsilon2, rmin2):
    """
    This **function** is used to calculate the A coefficient for Lorentz_Berthelot mix rule

    :param epsilon1: the epsilon parameter of the first atom
    :param rmin1: the rmin parameter of the first atom
    :param epsilon2: the epsilon parameter of the second atom
    :param rmin2: the rmin parameter of the second atom
    :return: the A coefficient for the atom pair
    """
    return np.sqrt(epsilon1 * epsilon2) * ((rmin1 + rmin2) ** 12)


def lorentz_berthelot_for_b(epsilon1, rmin1, epsilon2, rmin2):
    """
    This **function** is used to calculate the B coefficient for Lorentz_Berthelot mix rule

    :param epsilon1: the epsilon parameter of the first atom
    :param rmin1: the rmin parameter of the first atom
    :param epsilon2: the epsilon parameter of the second atom
    :param rmin2: the rmin parameter of the second atom
    :return: the B coefficient for the atom pair
    """
    return np.sqrt(epsilon1 * epsilon2) * 2 * ((rmin1 + rmin2) ** 6)


def _find_ab_lj(ljtypes, stat=True):
    """

    :param ljtypes:
    :param stat:
    :return:
    """
    coefficients_a = []
    coefficients_b = []

    for i, lj_single_i in enumerate(ljtypes):
        lj_i = LJType.get_type(lj_single_i + "-" + lj_single_i)
        if stat:
            j_max = len(ljtypes)
        else:
            j_max = i + 1
        for j in range(j_max):
            lj_j = LJType.get_type(ljtypes[j] + "-" + ljtypes[j])
            finded = False
            findnames = [lj_single_i + "-" + ljtypes[j], ljtypes[j] + "-" + lj_single_i]
            for findname in findnames:
                if findname in LJType.get_all_types():
                    finded = True
                    lj_ij = LJType.get_type(findname)
                    coefficients_a.append(
                        LJType.combining_method_A(lj_ij.epsilon, lj_ij.rmin, lj_ij.epsilon, lj_ij.rmin))
                    coefficients_b.append(
                        LJType.combining_method_B(lj_ij.epsilon, lj_ij.rmin, lj_ij.epsilon, lj_ij.rmin))
                    break
            if not finded:
                coefficients_a.append(LJType.combining_method_A(lj_i.epsilon, lj_i.rmin, lj_j.epsilon, lj_j.rmin))
                coefficients_b.append(LJType.combining_method_B(lj_i.epsilon, lj_i.rmin, lj_j.epsilon, lj_j.rmin))
    return coefficients_a, coefficients_b


def _get_checks(ljtypes, coefficients_a, coefficients_b):
    """

    :param ljtypes:
    :param coefficients_a:
    :param coefficients_b:
    :return:
    """
    checks = Xdict()
    count = 0
    for i in range(len(ljtypes)):
        check_string_a = ""
        check_string_b = ""
        for _ in range(len(ljtypes)):
            check_string_a += "%16.7e" % coefficients_a[count] + " "
            check_string_b += "%16.7e" % coefficients_b[count] + " "
            count += 1

        checks[i] = check_string_a + check_string_b
    return checks


def _judge_same_type(ljtypes, checks):
    """

    :param ljtypes:
    :param checks:
    :return:
    """
    same_type = {i: i for i in range(len(ljtypes))}
    for i in range(len(ljtypes) - 1, -1, -1):
        for j in range(i + 1, len(ljtypes)):
            if checks[i] == checks[j]:
                same_type[j] = i
    return same_type


def _get_real_lj(ljtypes, same_type):
    """

    :param ljtypes:
    :param same_type:
    :return:
    """
    real_ljtypes = []
    tosub = 0
    for i, lj_single_i in enumerate(ljtypes):
        if same_type[i] == i:
            real_ljtypes.append(lj_single_i)
            same_type[i] -= tosub
        else:
            same_type[i] = same_type[same_type[i]]
            tosub += 1
    return real_ljtypes


def write_lj(self):
    """
    This **function** is used to write SPONGE input file

    :param self: the Molecule instance
    :return: the string to write
    """
    ljtypes = []
    ljtypemap = Xdict()
    for atom in self.atoms:
        if atom.LJtype not in ljtypemap.keys():
            ljtypemap[atom.LJtype] = len(ljtypes)
            ljtypes.append(atom.LJtype)

    coefficients_a, coefficients_b = _find_ab_lj(ljtypes)

    checks = _get_checks(ljtypes, coefficients_a, coefficients_b)

    same_type = _judge_same_type(ljtypes, checks)

    real_ljtypes = _get_real_lj(ljtypes, same_type)

    real_as, real_bs = _find_ab_lj(real_ljtypes, False)

    towrite = "%d %d\n\n" % (len(self.atoms), len(real_ljtypes))
    count = 0
    for i in range(len(real_ljtypes)):
        for _ in range(i + 1):
            towrite += "%16.7e" % real_as[count] + " "
            count += 1
        towrite += "\n"
    towrite += "\n"

    count = 0
    for i in range(len(real_ljtypes)):
        for _ in range(i + 1):
            towrite += "%16.7e" % real_bs[count] + " "
            count += 1
        towrite += "\n"
    towrite += "\n"
    towrite += "\n".join(["%d" % (same_type[ljtypemap[atom.LJtype]]) for atom in self.atoms])
    return towrite


Molecule.Set_Save_SPONGE_Input("LJ")(write_lj)


#pylint: disable=unused-argument
def lj_todo(self, sys_kwarg, ene_kwarg, use_pbc):
    """
    This **function** is used to get MindSponge system and energy

    :param self: the Molecule instance
    :param sys_kwarg: a dict, the kwarg for system
    :param ene_kwarg: a dict, the kwarg for force field
    :return: None
    """
    from mindsponge.potential import LennardJonesEnergy
    if "lj" not in ene_kwarg:
        ene_kwarg["lj"] = Xdict()
        ene_kwarg["lj"]["function"] = lambda system, ene_kwarg: LennardJonesEnergy(
            epsilon=ene_kwarg["lj"]["epsilon"], use_pbc=use_pbc,
            sigma=ene_kwarg["lj"]["sigma"],
            length_unit='A', energy_unit='kcal/mol')
        ene_kwarg["lj"]["epsilon"] = []
        ene_kwarg["lj"]["sigma"] = []
    ljtypes = [LJType.get_type(atom.LJtype + "-" + atom.LJtype) for atom in self.atoms]
    ene_kwarg["lj"]["epsilon"].append([ljtype.epsilon for ljtype in ljtypes])
    ene_kwarg["lj"]["sigma"].append([ljtype.rmin / (4 ** (1 / 12) / 2) for ljtype in ljtypes])


Molecule.Set_MindSponge_Todo("LJ")(lj_todo)

set_global_alternative_names()



"""
This **module** set the basic configuration for ff14sb
"""
from ...helper import source, Xprint, set_real_global_variable

source("....")
amber = source("...amber")

AtomType.New_From_String(
    """
    name mass    charge[e]      LJtype
    HW   1.008   0.52422        HW
    OW   16      0              OW
    EP   0      -1.04844        EPW
    """)

bond_base.BondType.New_From_String(r"""
name    k[kcal/mol·A^-2]   b[A]
OW-HW   553                0.87243313
HW-HW   553                1.37120510
""")

angle_base.AngleType.New_From_String(r"""
name        k       b
HW-OW-HW    0       0
OW-HW-HW    0       0
""")

lj_base.LJType.New_From_String(r"""
name    epsilon[kcal/mol]   rmin[A]
OW-OW   0.2128008130           1.777167268
HW-HW   0                   0
EPW-EPW 0                   0
""")

virtual_atom_base.VirtualType2.New_From_String(r"""
name   atom0    atom1   atom2   k1         k2
EP     -3       -2      -1      0.1477206  0.1477206
""")

load_mol2(os.path.join(AMBER_DATA_DIR, "opc.mol2"), as_template=True)

ResidueType.set_type("H2O", ResidueType.get_type("WAT"))
ResidueType.set_type("HOH", ResidueType.get_type("WAT"))

set_real_global_variable("H2O", ResidueType.get_type("WAT"))
set_real_global_variable("HOH", ResidueType.get_type("WAT"))

amber.load_parameters_from_frcmod("ions1lm_126_tip4pew.frcmod")
amber.load_parameters_from_frcmod("ionsjc_tip4pew.frcmod")
amber.load_parameters_from_frcmod("ions234lm_126_tip4pew.frcmod")

load_mol2(os.path.join(AMBER_DATA_DIR, "atomic_ions.mol2"), as_template=True)

Xprint("""Reference for opc:
1. Water:
  Saeed Izadi, Ramu Anandakrishnan, and Alexey V. Onufriev
    Building Water Models: A Different Approach
    The Journal of Physical Chemistry Letters 2014 5 (21), 3863-3871, 
    DOI: 10.1021/jz501780a

2. Li+, Na+, K+, Rb+, Cs+, F-, Cl-, Br-, I-:
  In Suk Joung and Thomas E. Cheatham
    Determination of Alkali and Halide Monovalent Ion Parameters for Use in Explicitly Solvated Biomolecular Simulations
    The Journal of Physical Chemistry B 2008 112 (30), 9020-9041
    DOI: 10.1021/jp8001614

3. Ag+, Tl+, Cu+:
  Pengfei Li, Lin Frank Song, and Kenneth M. Merz
    Systematic Parameterization of Monovalent Ions Employing the Nonbonded Model
    Journal of Chemical Theory and Computation 2015 11 (4), 1645-1657, 
    DOI: 10.1021/ct500918t
    
4. Divalent Ions(Ba2+, Mg2+...)
  Pengfei Li and Kenneth M. Merz
    Taking into Account the Ion-Induced Dipole Interaction in the Nonbonded Model of Ions
    Journal of Chemical Theory and Computation 2014 10 (1), 289-297
    DOI: 10.1021/ct400751u

5. Trivalent and Tetravalent Cations(Al3+, Fe3+, Hf4+...)
  Pengfei Li, Lin Frank Song, and Kenneth M. Merz
    Parameterization of Highly Charged Metal Ions Using the 12-6-4 LJ-Type Nonbonded Model in Explicit Water
    The Journal of Physical Chemistry B 2015 119 (3), 883-895
    DOI: 10.1021/jp505875v  
""")
# pylint:disable=undefined-variable



"""
This **package** sets the basic configuration for GLYCAM-06j
"""
from ....helper import source, Xprint

source("....")
amber = source("...amber")

amber.load_parameters_from_parmdat(os.path.join("glycam_06j", "GLYCAM_06j.dat"))

load_mol2(os.path.join(AMBER_DATA_DIR, "glycam_06j", "terminal.mol2"), as_template=True)

ROH = ResidueType.get_type("ROH")
ROH.head = "O1"
ROH.head_next = "HO1"
ROH.head_length = 1.3

OME = ResidueType.get_type("OME")
OME.head = "O"
OME.head_next = "CH3"
OME.head_length = 1.3


def set_head(res, n):
    """
    This **function** sets the head of the carbohydrate residue type to the n-th oxygen atom
    :param res:
    :param n:
    :return:
    """
    head_dihedral = -60 if (n <= 6 and res.name[-1] not in "DU") else -180
    res.head = f"O{n}"
    res.head_next = f"C{n}"
    res.tail_length = 1.4
    res.head_link_conditions.clear()
    res.head_link_conditions.append({"atoms": [f"C{n}", f"O{n}"], "parameter": 111 / 180 * np.pi})
    res.head_link_conditions.append({"atoms": [f"C{n-1}", f"C{n}", f"O{n}"], "parameter": head_dihedral / 180 * np.pi})


set_global_alternative_names(set_head)

Xprint("""Reference for GLYCAM-06j:
  Kirschner, K.N., Yongye, A.B., Tschampel, S.M., González-Outeiriño, J., Daniels, C.R., Foley, B.L. and Woods
    GLYCAM06: A generalizable biomolecular force field. Carbohydrates.
    The Journal of Computational Chemistry 2008 29, 622-655, 
    DOI: 10.1002/jcc.20820
""")
